"""Testing OneHotEncode pipeline stages."""

import pytest
import pandas as pd

from pdpipe.col_generation import OneHotEncode
from pdpipe import cq


def _one_categ_df():
    return pd.DataFrame([["USA"], ["UK"], ["Greece"]], [1, 2, 3], ["Born"])


def _one_categ_single_row_df():
    return pd.DataFrame([["Greece"]], [1], ["Born"])


def _one_categ_df_large():
    return pd.DataFrame(
        data=[["USA"], ["UK"], ["Greece"], ["USA"], ["USA"], ["UK"], ["UK"]],
        index=[1, 2, 3, 4, 5, 6, 7],
        columns=["Born"],
    )


@pytest.mark.onehotencode
def test_onehotencode_one():
    """Basic binning test."""
    df = _one_categ_df()
    onehotencode = OneHotEncode("Born")
    res_df = onehotencode(df, verbose=True)
    assert "Born" not in res_df.columns
    assert "Born_Greece" not in res_df.columns
    assert "Born_UK" in res_df.columns
    assert res_df["Born_UK"][1] == 0
    assert res_df["Born_UK"][2] == 1
    assert res_df["Born_UK"][3] == 0
    assert "Born_USA" in res_df.columns
    assert res_df["Born_USA"][1] == 1
    assert res_df["Born_USA"][2] == 0
    assert res_df["Born_USA"][3] == 0

    # check when fitted
    df2 = _one_categ_single_row_df()
    assert onehotencode.is_fitted
    res_df2 = onehotencode(df2, verbose=True)
    print(res_df2)
    assert "Born" not in res_df2.columns
    assert "Born_Greece" not in res_df2.columns
    assert "Born_UK" in res_df2.columns
    assert res_df2["Born_UK"][1] == 0
    assert "Born_USA" in res_df.columns
    assert res_df2["Born_USA"][1] == 0


@pytest.mark.onehotencode
def test_onehotencode_large():
    """Basic binning test."""
    df = _one_categ_df()
    onehotencode = OneHotEncode("Born")
    res_df = onehotencode(df, verbose=True)
    assert "Born" not in res_df.columns
    assert "Born_Greece" not in res_df.columns
    assert "Born_UK" in res_df.columns
    assert res_df["Born_UK"][1] == 0
    assert res_df["Born_UK"][2] == 1
    assert res_df["Born_UK"][3] == 0
    assert "Born_USA" in res_df.columns
    assert res_df["Born_USA"][1] == 1
    assert res_df["Born_USA"][2] == 0
    assert res_df["Born_USA"][3] == 0

    # check when fitted
    df2 = _one_categ_df_large()
    assert onehotencode.is_fitted
    res_df2 = onehotencode(df2, verbose=True)
    print(res_df2)
    assert len(res_df2) == 7
    assert "Born" not in res_df2.columns
    assert "Born_Greece" not in res_df2.columns
    assert res_df2["Born_UK"][3] == 0
    assert res_df2["Born_USA"][3] == 0
    assert "Born_UK" in res_df2.columns
    assert res_df2["Born_UK"][2] == 1
    assert "Born_USA" in res_df.columns
    assert res_df2["Born_USA"][1] == 1


@pytest.mark.onehotencode
def test_onehotencode_no_drop_first():
    """Basic binning test."""
    df = _one_categ_df()
    onehotencode = OneHotEncode("Born", drop_first=False)
    res_df = onehotencode(df, verbose=True)
    assert "Born" not in res_df.columns
    assert "Born_UK" in res_df.columns
    assert res_df["Born_UK"][1] == 0
    assert res_df["Born_UK"][2] == 1
    assert res_df["Born_UK"][3] == 0
    assert "Born_USA" in res_df.columns
    assert res_df["Born_USA"][1] == 1
    assert res_df["Born_USA"][2] == 0
    assert res_df["Born_USA"][3] == 0
    assert "Born_Greece" in res_df.columns
    assert res_df["Born_Greece"][1] == 0
    assert res_df["Born_Greece"][2] == 0
    assert res_df["Born_Greece"][3] == 1

    # check when fitted
    df2 = _one_categ_single_row_df()
    assert onehotencode.is_fitted
    res_df2 = onehotencode(df2, verbose=True)
    print(res_df2)
    assert "Born" not in res_df2.columns
    assert "Born_UK" in res_df2.columns
    assert res_df2["Born_UK"][1] == 0
    assert "Born_USA" in res_df.columns
    assert res_df2["Born_USA"][1] == 0
    assert "Born_Greece" in res_df2.columns
    assert res_df2["Born_Greece"][1] == 1


def _two_categ_df():
    return pd.DataFrame(
        data=[["USA", "Bob"], ["UK", "Jack"], ["Greece", "Yan"]],
        index=[1, 2, 3],
        columns=["Born", "Name"],
    )


def _two_categ_single_row_df():
    return pd.DataFrame([["Greece", "Bob"]], [1], ["Born", "Name"])


@pytest.mark.onehotencode
def test_onehotencode_two():
    """Basic binning test."""
    df = _two_categ_df()
    onehotencode = OneHotEncode()
    res_df = onehotencode(df)
    assert "Born" not in res_df.columns
    assert "Born_Greece" not in res_df.columns
    assert "Born_UK" in res_df.columns
    assert res_df["Born_UK"][1] == 0
    assert res_df["Born_UK"][2] == 1
    assert res_df["Born_UK"][3] == 0
    assert "Born_USA" in res_df.columns
    assert res_df["Born_USA"][1] == 1
    assert res_df["Born_USA"][2] == 0
    assert res_df["Born_USA"][3] == 0
    assert "Name" not in res_df.columns
    assert "Name_Bob" not in res_df.columns
    assert "Name_Jack" in res_df.columns
    assert res_df["Name_Jack"][1] == 0
    assert res_df["Name_Jack"][2] == 1
    assert res_df["Name_Jack"][3] == 0
    assert "Name_Yan" in res_df.columns
    assert res_df["Name_Yan"][1] == 0
    assert res_df["Name_Yan"][2] == 0
    assert res_df["Name_Yan"][3] == 1

    # check when fitted
    df2 = _two_categ_single_row_df()
    assert onehotencode.is_fitted
    res_df2 = onehotencode(df2, verbose=True)
    print(res_df2)
    assert "Born" not in res_df2.columns
    assert "Born_Greece" not in res_df2.columns
    assert "Born_UK" in res_df2.columns
    assert res_df2["Born_UK"][1] == 0
    assert "Born_USA" in res_df.columns
    assert res_df2["Born_USA"][1] == 0
    assert "Name" not in res_df.columns
    assert "Name_Bob" not in res_df.columns
    assert "Name_Jack" in res_df.columns
    assert res_df2["Name_Jack"][1] == 0
    assert "Name_Yan" in res_df.columns
    assert res_df2["Name_Yan"][1] == 0


@pytest.mark.onehotencode
def test_onehotencode_one_with_exclude():
    """Basic binning test."""
    df = _two_categ_df()
    onehotencode = OneHotEncode(exclude_columns=["Name"])
    print(onehotencode._col_arg)
    res_df = onehotencode(df)
    assert "Born" not in res_df.columns
    assert "Name" in res_df.columns
    assert "Name_Bob" not in res_df.columns
    assert "Name_Jack" not in res_df.columns
    assert "Name_Yan" not in res_df.columns
    assert "Greece" not in res_df.columns
    assert "Born_UK" in res_df.columns
    assert res_df["Born_UK"][1] == 0
    assert res_df["Born_UK"][2] == 1
    assert res_df["Born_UK"][3] == 0
    assert "Born_USA" in res_df.columns
    assert res_df["Born_USA"][1] == 1
    assert res_df["Born_USA"][2] == 0
    assert res_df["Born_USA"][3] == 0

    # check when fitted
    df2 = _two_categ_single_row_df()
    assert onehotencode.is_fitted
    res_df2 = onehotencode(df2, verbose=True)
    print(res_df2)
    assert "Born" not in res_df2.columns
    assert "Born_Greece" not in res_df2.columns
    assert "Born_UK" in res_df2.columns
    assert res_df2["Born_UK"][1] == 0
    assert "Born_USA" in res_df.columns
    assert res_df2["Born_USA"][1] == 0
    assert "Name" in res_df.columns
    assert res_df2["Name"][1] == "Bob"
    assert "Name_Bob" not in res_df.columns
    assert "Name_Jack" not in res_df.columns
    assert "Name_Yan" not in res_df.columns


def _one_categ_df_with_nan():
    return pd.DataFrame([["USA"], ["UK"], [None]], [1, 2, 3], ["Born"])


@pytest.mark.onehotencode
def test_onehotencode_with_nan():
    """Basic binning test."""
    df = _one_categ_df_with_nan()
    onehotencode = OneHotEncode("Born")
    res_df = onehotencode(df)
    print(res_df)
    assert "Born" not in res_df.columns
    assert "Born_UK" not in res_df.columns
    assert "Born_nan" not in res_df.columns
    assert "Born_USA" in res_df.columns
    assert len(res_df.columns) == 1
    assert res_df["Born_USA"][1] == 1
    assert res_df["Born_USA"][2] == 0
    assert res_df["Born_USA"][3] == 0

    # check when fitted
    df2 = _one_categ_single_row_df()
    assert onehotencode.is_fitted
    res_df2 = onehotencode(df2, verbose=True)
    print(res_df2)
    assert "Born" not in res_df2.columns
    assert "Born_UK" not in res_df2.columns
    assert "Born_nan" not in res_df2.columns
    assert "Born_USA" in res_df.columns
    assert res_df2["Born_USA"][1] == 0


@pytest.mark.onehotencode
def test_onehotencode_with_dummy_na():
    """Basic binning test."""
    df = _one_categ_df_with_nan()
    onehotencode = OneHotEncode("Born", dummy_na=True)
    res_df = onehotencode(df)
    assert "Born" not in res_df.columns
    assert "Born_nan" not in res_df.columns
    assert "Born_UK" in res_df.columns
    assert res_df["Born_UK"][1] == 0
    assert res_df["Born_UK"][2] == 1
    assert res_df["Born_UK"][3] == 0
    assert "Born_USA" in res_df.columns
    assert res_df["Born_USA"][1] == 1
    assert res_df["Born_USA"][2] == 0
    assert res_df["Born_USA"][3] == 0

    # check when fitted
    df2 = _one_categ_single_row_df()
    assert onehotencode.is_fitted
    res_df2 = onehotencode(df2, verbose=True)
    print(res_df2)
    assert "Born" not in res_df2.columns
    assert "Born_nan" not in res_df2.columns
    assert "Born_USA" in res_df.columns
    assert res_df2["Born_USA"][1] == 0
    assert "Born_UK" in res_df2.columns
    assert res_df2["Born_UK"][1] == 0


@pytest.mark.onehotencode
def test_onehotencode_with_dummy_na_no_drop_first():
    """Basic binning test."""
    df = _one_categ_df_with_nan()
    onehotencode = OneHotEncode("Born", dummy_na=True, drop_first=False)
    res_df = onehotencode(df)
    assert "Born" not in res_df.columns
    assert "Born_UK" in res_df.columns
    assert res_df["Born_UK"][1] == 0
    assert res_df["Born_UK"][2] == 1
    assert res_df["Born_UK"][3] == 0
    assert "Born_USA" in res_df.columns
    assert res_df["Born_USA"][1] == 1
    assert res_df["Born_USA"][2] == 0
    assert res_df["Born_USA"][3] == 0
    assert "Born_nan" in res_df.columns
    assert res_df["Born_nan"][1] == 0
    assert res_df["Born_nan"][2] == 0
    assert res_df["Born_nan"][3] == 1

    # check when fitted
    df2 = _one_categ_single_row_df()
    assert onehotencode.is_fitted
    res_df2 = onehotencode(df2, verbose=True)
    print(res_df2)
    assert "Born" not in res_df2.columns
    assert "Born_Greece" not in res_df2.columns
    assert "Born_UK" in res_df2.columns
    assert res_df2["Born_UK"][1] == 0
    assert "Born_USA" in res_df.columns
    assert res_df2["Born_USA"][1] == 0
    assert "Born_nan" in res_df.columns
    assert res_df2["Born_nan"][1] == 0


@pytest.mark.onehotencode
def test_onehotencode_one_no_drop():
    """Basic binning test."""
    df = _one_categ_df()
    onehotencode = OneHotEncode("Born", drop=False)
    res_df = onehotencode(df, verbose=True)
    assert "Greece" not in res_df.columns
    assert "Born" in res_df.columns
    assert "Born_UK" in res_df.columns
    assert res_df["Born_UK"][1] == 0
    assert res_df["Born_UK"][2] == 1
    assert res_df["Born_UK"][3] == 0
    assert "Born_USA" in res_df.columns
    assert res_df["Born_USA"][1] == 1
    assert res_df["Born_USA"][2] == 0
    assert res_df["Born_USA"][3] == 0

    # check when fitted
    df2 = _one_categ_single_row_df()
    assert onehotencode.is_fitted
    res_df2 = onehotencode(df2, verbose=True)
    print(res_df2)
    assert "Born" in res_df2.columns
    assert res_df2["Born"][1] == "Greece"
    assert "Born_Greece" not in res_df2.columns
    assert "Born_UK" in res_df2.columns
    assert res_df2["Born_UK"][1] == 0
    assert "Born_USA" in res_df.columns
    assert res_df2["Born_USA"][1] == 0


@pytest.mark.onehotencode
def test_onehotencode_by_labels_cq():
    df = _two_categ_df()
    onehotencode = OneHotEncode(columns=cq.ByLabels(["Born", "Cat"]))
    res_df = onehotencode(df)
    assert "Born" not in res_df.columns
    assert "Born_Greece" not in res_df.columns
    assert "Born_UK" in res_df.columns
    assert res_df["Born_UK"][1] == 0
    assert res_df["Born_UK"][2] == 1
    assert res_df["Born_UK"][3] == 0
    assert "Born_USA" in res_df.columns
    assert res_df["Born_USA"][1] == 1
    assert res_df["Born_USA"][2] == 0
    assert res_df["Born_USA"][3] == 0
    assert "Name" in res_df.columns
    assert "Name_Bob" not in res_df.columns
    assert "Name_Jack" not in res_df.columns
    assert "Name_Yan" not in res_df.columns

    # check when fitted
    df2 = _two_categ_single_row_df()
    assert onehotencode.is_fitted
    res_df2 = onehotencode(df2, verbose=True)
    print(res_df2)
    assert "Born" not in res_df2.columns
    assert "Born_Greece" not in res_df2.columns
    assert "Born_UK" in res_df2.columns
    assert res_df2["Born_UK"][1] == 0
    assert "Born_USA" in res_df.columns
    assert res_df2["Born_USA"][1] == 0
    assert "Name" in res_df.columns
    assert "Name_Bob" not in res_df.columns
    assert "Name_Jack" not in res_df.columns
    assert "Name_Yan" not in res_df.columns


@pytest.mark.onehotencode
@pytest.mark.parametrize("verbose", [True, False])
def test_onehotencode_one_with_drop_first_colname(verbose):
    df = _one_categ_df()
    onehotencode = OneHotEncode("Born", drop_first="UK")
    res_df = onehotencode(df, verbose=verbose)
    assert "Born" not in res_df.columns
    assert "Born_UK" not in res_df.columns
    assert "Born_Greece" in res_df.columns
    assert res_df["Born_Greece"][1] == 0
    assert res_df["Born_Greece"][2] == 0
    assert res_df["Born_Greece"][3] == 1
    assert "Born_USA" in res_df.columns
    assert res_df["Born_USA"][1] == 1
    assert res_df["Born_USA"][2] == 0
    assert res_df["Born_USA"][3] == 0

    # check when fitted
    df2 = _one_categ_single_row_df()
    assert onehotencode.is_fitted
    res_df2 = onehotencode(df2, verbose=True)
    print(res_df2)
    assert "Born" not in res_df2.columns
    assert "Born_UK" not in res_df2.columns
    assert "Born_Greece" in res_df2.columns
    assert res_df2["Born_Greece"][1] == 1
    assert "Born_USA" in res_df.columns
    assert res_df2["Born_USA"][1] == 0



"""Fittable conditions for pdpipe.

In `pdipe`, pipeline stages have two optional constructor parameters that
accept callables that are treated as conditions: `prec` and `skip`. Both assume
input callables can accept a pandas.Dataframe object as input and return either
True or False. `prec` - representing the stage's precondition - determines
whether a stage *can* be applied to an input dataframe, while `skip` -
representing the stage's skip condition - determines whether it *should* be
applied. Accordingly, a stage throws a `FailedPreconditionError` if its
precondition is not statisfied, while it is skipped if its skip-condition is
not statisfied.

This module - `pdpipe.cond` - provides a way to easily generate `Condition`
objects, which are callable, and can easily be made fittable - to have their
result determined in fit time and preserved for future transforms - by
assigning the constructor parameter `fittable=True`. This enables the creation
of pipeline stages whose their effective inclusion in the pipeline is
determinedonly  when `fit_transform` is called; for example, whether
dimensionality reduction is required - once this decision is done in training
time it should be maintained for all future transforms of data (in test and
validation sets or in production).

Conditions objects also support the &, ^ and | binary operators - representing
boolean and, xor and or, respectively - and the ~ unary operator - representing
the boolean not operator.

So, for example, to get a condition that is satisfied by dataframes that are
missing at least one column from a list of column labels. one can use:

    >>> import pandas as pd; import pdpipe as pdp;
    >>> df = pd.DataFrame(
    ...    [[8,'a',5],[5,'b',7]], [1,2], ['num', 'chr', 'nur'])
    >>> cond = ~ pdp.cond.HasAllColumns(['num', 'chr'])
    >>> cond(df)
    False
    >>> cond = ~ pdp.cond.HasAllColumns(['num','go'])
    >>> cond(df)
    True

Similarly, to get a condition that is satisfied by dataframes that both has
columns names 'foo' and 'bar' AND has no missing values.

    >>> import pandas as pd; import pdpipe as pdp;
    >>> df = pd.DataFrame([[8, None],[5, 2]], [1,2], ['foo', 'bar'])
    >>> col_cond = pdp.cond.HasAllColumns(['foo', 'bar'])
    >>> missing_cond = pdp.cond.HasNoMissingValues()
    >>> (col_cond | missing_cond)(df)
    True
    >>> (col_cond & missing_cond)(df)
    False
    >>> df = pd.DataFrame([[8, 9],[5, 2]], [1,2], ['foo', 'bar'])
    >>> (col_cond & missing_cond)(df)
    True

While the same code but with XOR will yield the opposite results:

    >>> import pandas as pd; import pdpipe as pdp;
    >>> df = pd.DataFrame([[8, None],[5, 2]], [1,2], ['foo', 'bar'])
    >>> col_cond = pdp.cond.HasAllColumns(['foo', 'bar'])
    >>> missing_cond = pdp.cond.HasNoMissingValues()
    >>> (col_cond ^ missing_cond)(df)
    True
    >>> df = pd.DataFrame([[8, 9],[5, 2]], [1,2], ['foo', 'bar'])
    >>> (col_cond ^ missing_cond)(df)
    False
"""

from .shared import _list_str


class UnfittedConditionError(Exception):
    """An exception raised when a (non-fit) transform is attempted with an
    unfitted condition.
    """


class Condition(object):
    """A fittable condition that returns a boolean value from a dataframe.

    Parameters
    ----------
    func : callable
        A callable that given an input pandas.DataFrame objects returns a
        boolean value.
    fittable : bool, default False
        If set to True, this condition becomes fittable, and `func` is not
        called on calls of `transform()` of a fitted object. If set to False,
        the default, `func` is called on every call to transform. False by
        default.

    Example
    -------
        >>> import numpy as np; import pdpipe as pdp;
        >>> cond = pdp.cond.Condition(lambda df: 'a' in df.columns)
        >>> cond
        <pdpipe.Condition: By function>
        >>> col_drop = pdp.ColDrop(['lbl'], prec=cond)
    """

    def __init__(self, func, fittable=None):
        self._func = func
        self._fittable = fittable

    def __call__(self, df):
        """Returns column labels of qualified columns from an input dataframe.

        Parameters
        ----------
        df : pandas.DataFrame
            The input dataframe on which the condition is checked.

        Returns
        -------
        bool
            Either True of False.
        """
        try:
            return self.transform(df)
        except UnfittedConditionError:
            return self.fit_transform(df)

    def fit_transform(self, df):
        """Fits this condition and returns the result.

        Parameters
        ----------
        df : pandas.DataFrame
            The input dataframe on which the condition is checked.

        Returns
        -------
        bool
            Either True of False.
        """
        self._result = self._func(df)
        return self._result

    def fit(self, df):
        """Fits this condition on the input dataframe.

        Parameters
        ----------
        df : pandas.DataFrame
            The input dataframe on which the condition is checked.
        """
        self.fit_transform(df)

    def transform(self, df):
        """Returns the result of this condition.

        Is this Condition is fittable, it will return the result that was
        determined when fitted, if it's fitted, and throw an exception
        if it is not.

        Parameters
        ----------
        df : pandas.DataFrame
            The input dataframe on which the condition is checked.

        Returns
        -------
        bool
            Either True of False.
        """
        if not self._fittable:
            return self._func(df)
        try:
            return self._result
        except AttributeError:
            raise UnfittedConditionError

    def __repr__(self):
        fstr = ''
        if self._func.__doc__:  # pragma: no cover
            fstr = f' - {self._func.__doc__}'
        return f"<pdpipe.Condition: By function{fstr}>"

    # --- overriding boolean operators ---

    # need this because inner-scope functions aren't pickle-able
    class _AndCondition(object):

        def __init__(self, first, second):
            self.first = first
            self.second = second

        def __call__(self, df):
            return self.first(df) and self.second(df)

    def __and__(self, other):
        try:
            _func = Condition._AndCondition(self._func, other._func)
            _func.__doc__ = (
                f"{self._func.__doc__ or 'Anonymous condition 1'} AND "
                f"{other._func.__doc__ or 'Anonymous condition 2'}"
            )
            return Condition(func=_func)
        except AttributeError:
            return NotImplemented

    class _XorCondition(object):

        def __init__(self, first, second):
            self.first = first
            self.second = second

        def __call__(self, df):
            return self.first(df) != self.second(df)

    def __xor__(self, other):
        try:
            _func = Condition._XorCondition(self._func, other._func)
            _func.__doc__ = (
                f"{self._func.__doc__ or 'Anonymous condition 1'} XOR "
                f"{other._func.__doc__ or 'Anonymous condition 2'}"
            )
            return Condition(func=_func)
        except AttributeError:
            return NotImplemented

    class _OrCondition(object):

        def __init__(self, first, second):
            self.first = first
            self.second = second

        def __call__(self, df):
            return self.first(df) or self.second(df)

    def __or__(self, other):
        try:
            _func = Condition._OrCondition(self._func, other._func)
            _func.__doc__ = (
                f"{self._func.__doc__ or 'Anonymous condition 1'} OR "
                f"{other._func.__doc__ or 'Anonymous condition 2'}"
            )
            return Condition(func=_func)
        except AttributeError:
            return NotImplemented

    class _NotCondition(object):

        def __init__(self, first):
            self.first = first

        def __call__(self, df):
            return not self.first(df)

    def __invert__(self):
        _func = Condition._NotCondition(self._func)
        _func.__doc__ = f"NOT {self._func.__doc__ or 'Anonymous condition'}"
        return Condition(func=_func)


class PerColumnCondition(Condition):
    """Checks whether the columns of input dataframes statisfy a condition set.

    Parameters
    ----------
    conditions : callable or list-like
        The condition, or set of conditions, that columns of input dataframes
        must satisfy. Conditions are callables that accept a `pandas.Series`
        object and return a `bool` value.
    conditions_reduce : str, default 'all'
        How condition statisfaction results are reduced per-column, in case of
        multiple conditions. 'all' requires a column to satisfy all conditions,
        while 'any' requires at least one condition to be satisfied.
    columns_reduce : str, default 'all'
        How condition satisfaction results are reduced among multiple columns.
        'all' requires all columns of input dataframes to satisfy the given
        condition (in the case of multiple conditions, behaviour is determined
        by the `condition_reduce` parameter), while 'any' requires at least one
        column to statisfy it.
    **kwargs
        Additionaly accepts all keyword arguments of the constructor of
        Condition. See the documentation of Condition for details.

    Example
    -------
        >>> import pandas as pd; import pdpipe as pdp; import numpy as np;
        >>> df = pd.DataFrame(
        ...    [[8,'a',5],[5,'b',7]], [1,2], ['num', 'chr', 'nur'])
        >>> cond = pdp.cond.PerColumnCondition(
        ...     conditions=lambda x: x.dtype == np.int64,
        ... )
        >>> cond
        <pdpipe.Condition: Dataframes with all columns stasifying all \
conditions: anonymous condition>
        >>> cond(df)
        False
        >>> cond = pdp.cond.PerColumnCondition(
        ...     conditions=lambda x: x.dtype == np.int64,
        ...     columns_reduce='any',
        ... )
        >>> cond(df)
        True
        >>> cond = pdp.cond.PerColumnCondition(
        ...     conditions=[
        ...         lambda x: x.dtype == np.int64,
        ...         lambda x: x.dtype == object,
        ...     ],
        ... )
        >>> cond(df)
        False
        >>> cond = pdp.cond.PerColumnCondition(
        ...     conditions=[
        ...         lambda x: x.dtype == np.int64,
        ...         lambda x: x.dtype == object,
        ...     ],
        ...     conditions_reduce='any',
        ... )
        >>> cond(df)
        True
    """

    class _ConditionFunction(object):

        def __init__(self, conditions, cond_reduce, col_reduce):
            self.conditions = conditions
            self.cond_reduce = cond_reduce
            self.col_reduce = col_reduce

        def __call__(self, df):
            return self.col_reduce([
                self.cond_reduce([
                    cond(df[lbl])
                    for cond in self.conditions
                ])
                for lbl in df.columns
            ])

    def __init__(self, conditions, conditions_reduce=None, columns_reduce=None,
                 **kwargs):
        # handling default args and input types
        if not hasattr(conditions, '__iter__'):
            conditions = [conditions]
        if conditions_reduce is None:
            conditions_reduce = 'all'
        if columns_reduce is None:
            columns_reduce = 'all'
        # building class attributes
        self._conditions = conditions
        self._cond_reduce_str = conditions_reduce
        self._col_reduce_str = columns_reduce
        self._conditions_str = ', '.join([
            c.__doc__ or 'anonymous condition'
            for c in conditions
        ])
        if conditions_reduce == 'all':
            self._cond_reduce = all
        elif conditions_reduce == 'any':
            self._cond_reduce = any
        else:
            raise ValueError((
                "The only valid arguments to the `conditions_reduce` parameter"
                " of PerColumnCondition are 'all' and 'any'!"
            ))
        if columns_reduce == 'all':
            self._col_reduce = all
        elif columns_reduce == 'any':
            self._col_reduce = any
        else:
            raise ValueError((
                "The only valid arguments to the `columns_reduce` parameter"
                " of PerColumnCondition are 'all' and 'any'!"
            ))
        # building resulting function
        _func = PerColumnCondition._ConditionFunction(
            conditions=self._conditions,
            cond_reduce=self._cond_reduce,
            col_reduce=self._col_reduce,
        )
        doc_str = "Dataframes with {} columns stasifying {} conditions: {}"
        self._func_doc = doc_str.format(
            self._col_reduce_str, self._cond_reduce_str, self._conditions_str)
        _func.__doc__ = self._func_doc
        kwargs['func'] = _func
        super().__init__(**kwargs)

    def __repr__(self):
        return f"<pdpipe.Condition: {self._func_doc}>"


class HasAllColumns(Condition):
    """Checks whether input dataframes contain a list of columns.

    Parameters
    ----------
    labels : single label or list-like
        Column labels to check for.
    **kwargs
        Additionaly accepts all keyword arguments of the constructor of
        Condition. See the documentation of Condition for details.

    Example
    -------
        >>> import pandas as pd; import pdpipe as pdp;
        >>> df = pd.DataFrame(
        ...    [[8,'a',5],[5,'b',7]], [1,2], ['num', 'chr', 'nur'])
        >>> cond = pdp.cond.HasAllColumns('num')
        >>> cond
        <pdpipe.Condition: Has all columns in num>
        >>> cond(df)
        True
        >>> cond = pdp.cond.HasAllColumns(['num', 'chr'])
        >>> cond(df)
        True
        >>> cond = pdp.cond.HasAllColumns(['num', 'gar'])
        >>> cond(df)
        False
    """

    def __init__(self, labels, **kwargs):
        if isinstance(labels, str) or not hasattr(labels, '__iter__'):
            labels = [labels]
        self._labels = labels
        self._labels_str = _list_str(self._labels)
        def _func(df):  # noqa: E306
            return all([
                lbl in df.columns
                for lbl in self._labels
            ])
        _func.__doc__ = f"Dataframes with colums {self._labels_str}"
        kwargs['func'] = _func
        super().__init__(**kwargs)

    def __repr__(self):
        return f"<pdpipe.Condition: Has all columns in {self._labels_str}>"


class ColumnsFromList(PerColumnCondition):
    """Checks whether input dataframes contain columns from a list.

    Parameters
    ----------
    labels : single label or list-like
        Column labels to check for.
    columns_reduce : str, default 'all'
        How condition satisfaction results are reduced among multiple columns.
        'all' requires all columns of input dataframes to satisfy the given
        condition, while 'any' requires at least one column to statisfy it.
    **kwargs
        Additionaly accepts all keyword arguments of the constructor of
        Condition. See the documentation of Condition for details.

    Example
    -------
        >>> import pandas as pd; import pdpipe as pdp;
        >>> df = pd.DataFrame(
        ...    [[8,'a',5],[5,'b',7]], [1,2], ['num', 'chr', 'nur'])
        >>> cond = pdp.cond.ColumnsFromList('num')
        >>> cond
        <pdpipe.Condition: Dataframes with all columns stasifying all \
conditions: Series with labels in num>
        >>> cond(df)
        False
        >>> cond = pdp.cond.ColumnsFromList(['num', 'chr', 'nur'])
        >>> cond(df)
        True
        >>> cond = pdp.cond.ColumnsFromList(
        ...     ['num', 'gar'], columns_reduce='any')
        >>> cond(df)
        True
    """

    class _SeriesLblCondition(object):

        def __init__(self, labels):
            self.labels = labels

        def __call__(self, series):
            return series.name in self.labels

    def __init__(self, labels, columns_reduce=None, **kwargs):
        if isinstance(labels, str) or not hasattr(labels, '__iter__'):
            labels = [labels]
        self._labels = labels
        self._labels_str = _list_str(self._labels)
        _func = ColumnsFromList._SeriesLblCondition(self._labels)
        _func.__doc__ = f"Series with labels in {self._labels_str}"
        kwargs['conditions'] = [_func]
        kwargs['columns_reduce'] = columns_reduce
        super().__init__(**kwargs)


class HasNoColumn(Condition):
    """Checks whether input dataframes contains no column from a list.

    Parameters
    ----------
    labels : single label or list-like
        Column labels to check for.
    **kwargs
        Additionaly accepts all keyword arguments of the constructor of
        Condition. See the documentation of Condition for details.

    Example
    -------
        >>> import pandas as pd; import pdpipe as pdp;
        >>> df = pd.DataFrame(
        ...    [[8,'a',5],[5,'b',7]], [1,2], ['num', 'chr', 'nur'])
        >>> cond = pdp.cond.HasNoColumn('num')
        >>> cond
        <pdpipe.Condition: Has no column in num>
        >>> cond(df)
        False
        >>> cond = pdp.cond.HasNoColumn(['num', 'gar'])
        >>> cond(df)
        False
        >>> cond = pdp.cond.HasNoColumn(['ph', 'gar'])
        >>> cond(df)
        True
    """

    class _NoColumnsFunc(object):

        def __init__(self, labels):
            self.labels = labels

        def __call__(self, df):
            return all([
                lbl not in df.columns
                for lbl in self.labels
            ])

    def __init__(self, labels, **kwargs):
        if isinstance(labels, str) or not hasattr(labels, '__iter__'):
            labels = [labels]
        self._labels = labels
        self._labels_str = _list_str(self._labels)
        _func = HasNoColumn._NoColumnsFunc(self._labels)
        _func.__doc__ = f"Dataframes with no colum from {self._labels_str}"
        kwargs['func'] = _func
        super().__init__(**kwargs)

    def __repr__(self):
        return f"<pdpipe.Condition: Has no column in {self._labels_str}>"


class HasAtMostMissingValues(Condition):
    """Checks whether input dataframes has no more than X missing values.

    Parameters
    ----------
    n_missing : int or float
        If int, then interpreted as the maximal allowed number of missing
        values in input dataframes. If float, interpreted as the maximal
        allowed ratio of missing values in input dataframes.
    **kwargs
        Additionally accepts all keyword arguments of the constructor of
        Condition. See the documentation of Condition for details.

    Example
    -------
        >>> import pandas as pd; import pdpipe as pdp;
        >>> df = pd.DataFrame(
        ...    [[None,'a',5],[5,None,7]], [1,2], ['num', 'chr', 'nur'])
        >>> cond = pdp.cond.HasAtMostMissingValues(1)
        >>> cond
        <pdpipe.Condition: Has at most 1 missing values>
        >>> cond(df)
        False
        >>> cond = pdp.cond.HasAtMostMissingValues(2)
        >>> cond(df)
        True
        >>> cond = pdp.cond.HasAtMostMissingValues(0.4)
        >>> cond(df)
        True
        >>> cond = pdp.cond.HasAtMostMissingValues(0.2)
        >>> cond(df)
        False
    """

    class _IntMissingValuesFunc(object):

        def __init__(self, n_missing):
            self.n_missing = n_missing

        def __call__(self, df):
            nmiss = df.isna().sum().sum()
            return nmiss <= self.n_missing

    class _FloatMissingValuesFunc(object):

        def __init__(self, n_missing):
            self.n_missing = n_missing

        def __call__(self, df):
            nmiss = df.isna().sum().sum()
            return (nmiss / df.size) <= self.n_missing

    def __init__(self, n_missing, **kwargs):
        self._n_missing = n_missing
        if isinstance(n_missing, int):
            _func = HasAtMostMissingValues._IntMissingValuesFunc(n_missing)
        elif isinstance(n_missing, float):
            _func = HasAtMostMissingValues._FloatMissingValuesFunc(n_missing)
        else:
            raise ValueError("n_missing should be of type int or float!")
        _func.__doc__ = (
            f"Dataframes with at most {self._n_missing} missing values"
        )
        kwargs['func'] = _func
        super().__init__(**kwargs)

    def __repr__(self):
        return f"<pdpipe.Condition: " \
               f"Has at most {self._n_missing} missing values>"


class HasNoMissingValues(HasAtMostMissingValues):
    """Checks whether input dataframes has no missing values.

    Parameters
    ----------
    **kwargs
        Accepts all keyword arguments of the constructor of Condition. See the
        documentation of Condition for details.

    Example
    -------
        >>> import pandas as pd; import pdpipe as pdp;
        >>> df = pd.DataFrame(
        ...    [[None,'a',5],[5,'b',7]], [1,2], ['num', 'chr', 'nur'])
        >>> cond = pdp.cond.HasNoMissingValues()
        >>> cond
        <pdpipe.Condition: Has no missing values>
        >>> cond(df)
        False
    """

    def __init__(self, **kwargs):
        kwargs['n_missing'] = 0
        super().__init__(**kwargs)

    def __repr__(self):
        return "<pdpipe.Condition: Has no missing values>"



"""Testing pdpipe util module."""

import pandas as pd
import pytest

from pdpipe.util import out_of_place_col_insert


def _test_df():
    return pd.DataFrame(
        data=[[1, 'a'], [2, 'b']],
        index=[1, 2],
        columns=['num', 'char']
    )


def test_out_of_place_col_insert_all_params():
    """Testing the ColDrop pipeline stage."""
    df = _test_df()
    series = pd.Series(
        data=[10, 20],
        index=[1, 2],
        name='tens')

    result_df = out_of_place_col_insert(df, series, 1, 'Tigers')
    assert 'tens' not in result_df.columns
    assert 'Tigers' in result_df.columns
    assert result_df.columns.get_loc('Tigers') == 1
    assert result_df['Tigers'][1] == 10
    assert result_df['Tigers'][2] == 20


def test_out_of_place_col_insert_no_col_name():
    """Testing the ColDrop pipeline stage."""
    df = _test_df()
    series = pd.Series(
        data=[10, 20],
        index=[1, 2],
        name='tens')

    result_df = out_of_place_col_insert(df, series, 1)
    assert 'tens' in result_df.columns
    assert result_df.columns.get_loc('tens') == 1
    assert result_df['tens'][1] == 10
    assert result_df['tens'][2] == 20


def test_out_of_place_col_insert_nameless_error():
    """Testing the ColDrop pipeline stage."""
    df = _test_df()
    series = pd.Series(
        data=[10, 20],
        index=[1, 2])

    with pytest.raises(ValueError):
        out_of_place_col_insert(df, series, 1)


def test_out_of_place_col_last_position():
    """Testing the ColDrop pipeline stage."""
    df = _test_df()
    series = pd.Series(
        data=[10, 20],
        index=[1, 2],
        name='tens')

    result_df = out_of_place_col_insert(df, series, len(df.columns), 'Tigers')
    assert 'tens' not in result_df.columns
    assert 'Tigers' in result_df.columns
    assert result_df.columns.get_loc('Tigers') == 2
    assert result_df['Tigers'][1] == 10
    assert result_df['Tigers'][2] == 20



"""Basic pdpipe PdPipelineStages."""

from typing import Optional, List, Dict
from collections import deque

import pandas
from strct.dicts import reverse_dict_partial

from pdpipe.core import PdPipelineStage, ColumnsBasedPipelineStage
# from pdpipe.util import out_of_place_col_insert
from pdpipe.shared import (
    _interpret_columns_param,
    _list_str
)

import pdpipe.cond as cond
from pdpipe.types import ColumnsParamType


class ColDrop(ColumnsBasedPipelineStage):
    """A pipeline stage that drops columns by name.

    Parameters
    ----------
    columns : single label, list-like or callable
        The label, or an iterable of labels, of columns to drop. Alternatively,
        this parameter can be assigned a callable returning an iterable of
        labels from an input pandas.DataFrame. See pdpipe.cq.
    errors : {‘ignore’, ‘raise’}, default ‘raise’
        If ‘ignore’, suppress error and existing labels are dropped.

    Example
    -------
        >>> import pandas as pd; import pdpipe as pdp;
        >>> df = pd.DataFrame([[8,'a'],[5,'b']], [1,2], ['num', 'char'])
        >>> pdp.ColDrop('num').apply(df)
          char
        1    a
        2    b
    """

    def __init__(
        self,
        columns: ColumnsParamType,
        errors: Optional[str] = None,
        **kwargs: object,
    ) -> None:
        self._errors = errors
        self._post_cond = cond.HasNoColumn(columns)
        super_kwargs = {
            'columns': columns,
            'desc_temp': 'Drop columns {}',
        }
        super_kwargs.update(**kwargs)
        super_kwargs['none_columns'] = 'error'
        super().__init__(**super_kwargs)

    def _prec(self, df: pandas.DataFrame) -> bool:
        if self._errors != 'ignore':
            return super()._prec(df)
        return True

    def _post(self, df: pandas.DataFrame) -> bool:
        return self._post_cond(df)

    def _transformation(
        self, df: pandas.DataFrame, verbose: bool, fit: bool,
    ) -> pandas.DataFrame:
        return df.drop(
            self._get_columns(df, fit=fit), axis=1, errors=self._errors)


class ValDrop(ColumnsBasedPipelineStage):
    """A pipeline stage that drops rows by value.

    Parameters
    ----------
    values : list-like
        A list of the values to drop.
    columns : single label, list-like or callable, default None
        The label, or an iterable of labels, of columns to check for the given
        values. Alternatively, this parameter can be assigned a callable
        returning an iterable of labels from an input pandas.DataFrame. See
        pdpipe.cq. If set to None, all columns are checked.
    exclude_columns : object, iterable or callable, optional
        The label, or an iterable of labels, of columns to exclude, given the
        `columns` parameter. Alternatively, this parameter can be assigned a
        callable returning a labels iterable from an input pandas.DataFrame.
        See pdpipe.cq. Optional. By default no columns are excluded.

    Example
    -------
        >>> import pandas as pd; import pdpipe as pdp;
        >>> df = pd.DataFrame([[1,4],[4,5],[18,11]], [1,2,3], ['a','b'])
        >>> pdp.ValDrop([4], 'a').apply(df)
            a   b
        1   1   4
        3  18  11
        >>> pdp.ValDrop([4]).apply(df)
            a   b
        3  18  11
    """

    def __init__(
        self,
        values: List[object],
        columns: ColumnsParamType = None,
        **kwargs: object,
    ) -> None:
        self._values = values
        self._values_str = _list_str(self._values)
        super_kwargs = {
            'columns': columns,
            'desc_temp': f'Drop values {self._values_str} in columns {{}}',
        }
        super_kwargs.update(**kwargs)
        super_kwargs['none_columns'] = 'all'
        super().__init__(**super_kwargs)

    def _transformation(
        self, df: pandas.DataFrame, verbose: bool, fit: bool,
    ) -> pandas.DataFrame:
        inter_df = df
        before_count = len(inter_df)
        columns_to_check = self._get_columns(df, fit=fit)
        for col in columns_to_check:
            inter_df = inter_df[~inter_df[col].isin(self._values)]
        if verbose:
            print(f"{before_count - len(inter_df)} rows dropped.")
        return inter_df


class ValKeep(ColumnsBasedPipelineStage):
    """A pipeline stage that keeps rows by value.

    Parameters
    ----------
    values : list-like
        A list of the values to keep.
    columns : single label, list-like or callable, default None
        The label, or an iterable of labels, of columns to check for the given
        values. Alternatively, this parameter can be assigned a callable
        returning an iterable of labels from an input pandas.DataFrame. See
        pdpipe.cq. If set to None, all columns are checked.
    exclude_columns : object, iterable or callable, optional
        The label, or an iterable of labels, of columns to exclude, given the
        `columns` parameter. Alternatively, this parameter can be assigned a
        callable returning a labels iterable from an input pandas.DataFrame.
        See pdpipe.cq. Optional. By default no columns are excluded.

    Example
    -------
        >>> import pandas as pd; import pdpipe as pdp;
        >>> df = pd.DataFrame([[1,4],[4,5],[5,11]], [1,2,3], ['a','b'])
        >>> pdp.ValKeep([4, 5], 'a').apply(df)
           a   b
        2  4   5
        3  5  11
        >>> pdp.ValKeep([4, 5]).apply(df)
           a  b
        2  4  5
    """

    def __init__(self, values, columns=None, **kwargs):
        self._values = values
        self._values_str = _list_str(self._values)
        super_kwargs = {
            'columns': columns,
            'desc_temp': f'Keep values {self._values_str} in columns {{}}',
        }
        super_kwargs.update(**kwargs)
        super_kwargs['none_columns'] = 'all'
        super().__init__(**super_kwargs)

    def _transformation(self, df, verbose, fit):
        inter_df = df
        before_count = len(inter_df)
        columns_to_check = self._get_columns(df, fit=fit)
        for col in columns_to_check:
            inter_df = inter_df[inter_df[col].isin(self._values)]
        if verbose:
            print(f"{before_count - len(inter_df)} rows dropped.")
        return inter_df


class ColRename(PdPipelineStage):
    """A pipeline stage that renames a column or columns.

    Parameters
    ----------
    rename_mapper : dict-like or function
        Maps old column names to new ones.

    Example
    -------
        >>> import pandas as pd; import pdpipe as pdp;
        >>> df = pd.DataFrame([[8,'a'],[5,'b']], [1,2], ['num', 'char'])
        >>> pdp.ColRename({'num': 'len', 'char': 'initial'}).apply(df)
           len initial
        1    8       a
        2    5       b
    """

    _DEF_COLDRENAME_EXC_MSG = ("ColRename stage failed because not all columns"
                               " {} were found in input dataframe.")

    def __init__(self, rename_mapper, **kwargs):
        self._rename_mapper = rename_mapper
        try:
            columns_str = _list_str(list(rename_mapper.keys()))
            mapper_repr = str(rename_mapper)
            keys_set = set(self._rename_mapper.keys())

            def _tprec(df):
                return keys_set.issubset(df.columns)
        except AttributeError:
            mapper_repr = rename_mapper.__name__
            doc = rename_mapper.__doc__
            if doc is None:
                columns_str = f"by func {rename_mapper.__name__}"
            else:
                columns_str = (
                    f"by func {rename_mapper.__name__} with "
                    f"doc: {rename_mapper.__doc__}"
                )

            def _tprec(df):
                return True
        try:
            suffix = 's' if len(rename_mapper) > 1 else ''
        except TypeError:
            suffix = 's'
        self._tprec = _tprec
        super_kwargs = {
            'exmsg': ColRename._DEF_COLDRENAME_EXC_MSG.format(columns_str),
            'desc': f"Rename column{suffix} with {mapper_repr}",
        }
        super_kwargs.update(**kwargs)
        super().__init__(**super_kwargs)

    def _prec(self, df):
        return self._tprec(df)

    def _transform(self, df, verbose):
        return df.rename(columns=self._rename_mapper)


class DropNa(PdPipelineStage):
    """A pipeline stage that drops null values.

    Supports all parameter supported by pandas.dropna function.

    Example
    -------
        >>> import pandas as pd; import pdpipe as pdp;
        >>> df = pd.DataFrame([[1,4],[4,None],[1,11]], [1,2,3], ['a','b'])
        >>> pdp.DropNa().apply(df)
           a     b
        1  1   4.0
        3  1  11.0
    """

    _DEF_DROPNA_EXC_MSG = "DropNa stage failed."
    _DEF_DROPNA_APP_MSG = "Dropping null values..."
    _DROPNA_KWARGS = ['axis', 'how', 'thresh', 'subset', 'inplace']

    def __init__(self, **kwargs):
        common = set(kwargs.keys()).intersection(DropNa._DROPNA_KWARGS)
        self.dropna_kwargs = {key: kwargs.pop(key) for key in common}
        super_kwargs = {
            'exmsg': DropNa._DEF_DROPNA_EXC_MSG,
            'desc': "Drops null values."
        }
        super_kwargs.update(**kwargs)
        super().__init__(**super_kwargs)

    def _prec(self, df):
        return True

    def _transform(self, df, verbose):
        before_count = len(df)
        ncols_before = len(df.columns)
        inter_df = df.dropna(**self.dropna_kwargs)
        if verbose:
            print(
                f"{before_count - len(inter_df)} rows, "
                f"{ncols_before - len(inter_df.columns)} columns dropeed"
            )
        return inter_df


class SetIndex(PdPipelineStage):
    """A pipeline stage that set existing columns as index.

    Supports all parameter supported by pandas.set_index function except for
    `inplace`.

    Example
    -------
        >> import pandas as pd; import pdpipe as pdp;
        >> df = pd.DataFrame([[1,4],[3, 11]], [1,2], ['a','b'])
        >> pdp.SetIndex('a').apply(df)
            b
        a
        1   4
        3  11
    """

    _DEF_SETIDX_EXC_MSG = "SetIndex stage failed."
    _DEF_SETIDX_APP_MSG = "Setting indexes..."
    _SETINDEX_KWARGS = ['drop', 'append', 'verify_integrity']

    def __init__(self, keys, **kwargs):
        common = set(kwargs.keys()).intersection(SetIndex._SETINDEX_KWARGS)
        self.setindex_kwargs = {key: kwargs.pop(key) for key in common}
        self.keys = keys
        if hasattr(keys, '__iter__') and not isinstance(keys, str):
            def _tprec(df):
                return all([k in df.columns for k in keys])
        else:
            def _tprec(df):
                return keys in df.columns
        self._tprec = _tprec
        super_kwargs = {
            'exmsg': SetIndex._DEF_SETIDX_EXC_MSG,
            'desc': "Set indexes."
        }
        super_kwargs.update(**kwargs)
        super().__init__(**super_kwargs)

    def _prec(self, df):
        return self._tprec(df)

    def _transform(self, df, verbose):
        return df.set_index(keys=self.keys, **self.setindex_kwargs)


class FreqDrop(PdPipelineStage):
    """A pipeline stage that drops rows by value frequency.

    Parameters
    ----------
    threshold : int
        The minimum frequency required for a value to be kept.
    column : str
        The name of the colum to check for the given value frequency.

    Example
    -------
        >>> import pandas as pd; import pdpipe as pdp;
        >>> df = pd.DataFrame([[1,4],[4,5],[1,11]], [1,2,3], ['a','b'])
        >>> pdp.FreqDrop(2, 'a').apply(df)
           a   b
        1  1   4
        3  1  11
    """

    _DEF_FREQDROP_EXC_MSG = ("FreqDrop stage failed because column {} was not"
                             " found in input dataframe.")
    _DEF_FREQDROP_DESC = "Drop values with frequency < {} in column {}."

    def __init__(self, threshold, column, **kwargs):
        self._threshold = threshold
        self._column = column
        super_kwargs = {
            'exmsg': FreqDrop._DEF_FREQDROP_EXC_MSG.format(self._column),
            'desc': FreqDrop._DEF_FREQDROP_DESC.format(
                self._threshold, self._column)
        }
        super_kwargs.update(**kwargs)
        super().__init__(**super_kwargs)

    def _prec(self, df):
        return self._column in df.columns

    def _transform(self, df, verbose):
        inter_df = df
        before_count = len(inter_df)
        valcount = df[self._column].value_counts()
        to_drop = valcount[valcount < self._threshold].index
        inter_df = inter_df[~inter_df[self._column].isin(to_drop)]
        if verbose:
            print(f"{before_count - len(inter_df)} rows dropped.")
        return inter_df


class ColReorder(PdPipelineStage):
    """A pipeline stage that reorders columns.

    Parameters
    ----------
    positions : dict
        A mapping of column names to their desired positions after reordering.
        Columns not included in the mapping will maintain their relative
        positions over the non-mapped colums.

    Example
    -------
        >>> import pandas as pd; import pdpipe as pdp;
        >>> df = pd.DataFrame([[8,4,3,7]], columns=['a', 'b', 'c', 'd'])
        >>> pdp.ColReorder({'b': 0, 'c': 3}).apply(df)
           b  a  d  c
        0  4  8  7  3
    """

    _DEF_ORD_EXC_MSG = ("ColReorder stage failed because not all columns in {}"
                        " were found in input dataframe.")

    def __init__(self, positions, **kwargs):
        self._col_to_pos = positions
        self._pos_to_col = reverse_dict_partial(positions)
        super_kwargs = {
            'exmsg': ColReorder._DEF_ORD_EXC_MSG.format(self._col_to_pos),
            'desc': f"Reorder columns by {self._col_to_pos}",
        }
        super_kwargs.update(**kwargs)
        super().__init__(**super_kwargs)

    def _prec(self, df):
        return set(self._col_to_pos.keys()).issubset(df.columns)

    def _transform(self, df, verbose):
        cols = df.columns
        map_cols = list(self._col_to_pos.keys())
        non_map_cols = deque(x for x in cols if x not in map_cols)
        new_columns = []
        try:
            for pos in range(len(cols)):
                if pos in self._pos_to_col:
                    new_columns.append(self._pos_to_col[pos])
                else:
                    new_columns.append(non_map_cols.popleft())
            return df[new_columns]
        except (IndexError):
            raise ValueError(f"Bad positions mapping given: {new_columns}")


class RowDrop(ColumnsBasedPipelineStage):
    """A pipeline stage that drop rows by callable conditions.

    Parameters
    ----------
    conditions : list-like or dict
        The list of conditions that make a row eligible to be dropped. Each
        condition must be a callable that take a cell value and return a bool
        value. If a list of callables is given, the conditions are checked for
        each column value of each row. If a dict mapping column labels to
        callables is given, then each condition is only checked for the column
        values of the designated column.
    reduce : 'any', 'all' or 'xor', default 'any'
        Determines how row conditions are reduced. If set to 'all', a row must
        satisfy all given conditions to be dropped. If set to 'any', rows
        satisfying at least one of the conditions are dropped. If set to 'xor',
        rows satisfying exactly one of the conditions will be dropped. Set to
        'any' by default.
    columns : str or iterable, optional
        The label, or an iterable of labels, of columns. Alternatively,
        this parameter can be assigned a callable returning an iterable of
        labels from an input pandas.DataFrame. See pdpipe.cq. If given,
        input conditions will be applied to the sub-dataframe made up of
        these columns to determine which rows to drop. Ignored if `conditions`
        is provided with a dict object. If `conditions` is a list and this
        parameter is not provided, all columns are checked (unless
        `exclude_columns` is additionally provided)
    exclude_columns : object, iterable or callable, optional
        The label, or an iterable of labels, of columns to exclude, given the
        `columns` parameter. Alternatively, this parameter can be assigned a
        callable returning a labels iterable from an input pandas.DataFrame.
        See pdpipe.cq. Optional. By default no columns are excluded.

    Example
    -------
        >>> import pandas as pd; import pdpipe as pdp;
        >>> df = pd.DataFrame([[1,4],[4,5],[5,11]], [1,2,3], ['a','b'])
        >>> pdp.RowDrop([lambda x: x < 2]).apply(df)
           a   b
        2  4   5
        3  5  11
        >>> pdp.RowDrop({'a': lambda x: x == 4}).apply(df)
           a   b
        1  1   4
        3  5  11
    """

    _REDUCERS = {
        'all': all,
        'any': any,
        'xor': lambda x: sum(x) == 1
    }

    class DictRowCond(object):
        """Filter rows by a dict of conditions."""

        def __init__(self, conditions, reducer):
            self.conditions = conditions
            self.reducer = reducer

        def __call__(self, row):
            res = [cond(row[lbl]) for lbl, cond in self.conditions.items()]
            return self.reducer(res)

    class ListRowCond(object):
        """Filter rows by a list of conditions."""

        def __init__(self, conditions, reducer):
            self.conditions = conditions
            self.reducer = reducer

        def __call__(self, row):
            res = [self.reducer(row.apply(cond)) for cond in self.conditions]
            return self.reducer(res)

    def _row_condition_builder(self, conditions, reduce):
        reducer = RowDrop._REDUCERS[reduce]
        if self._cond_is_dict:
            row_cond = RowDrop.DictRowCond(
                conditions=conditions, reducer=reducer)
        else:
            row_cond = RowDrop.ListRowCond(
                conditions=conditions, reducer=reducer)
        return row_cond

    def __init__(self, conditions, reduce=None, columns=None, **kwargs):
        self._conditions = conditions
        if reduce is None:
            reduce = 'any'
        self._reduce = reduce
        if reduce not in RowDrop._REDUCERS.keys():
            raise ValueError((
                "{} is an unsupported argument for the 'reduce' parameter of "
                "the RowDrop constructor!").format(reduce))
        self._cond_is_dict = isinstance(conditions, dict)
        if self._cond_is_dict:
            valid = all([callable(cond) for cond in conditions.values()])
            if not valid:
                raise ValueError(
                    "Condition dicts given to RowDrop must map to callables!")
            columns = list(conditions.keys())
        else:
            valid = all([callable(cond) for cond in conditions])
            if not valid:
                raise ValueError(
                    "RowDrop condition lists can contain only callables!")
        self._row_cond = self._row_condition_builder(conditions, reduce)
        super_kwargs = {
            'columns': columns,
            'desc_temp': 'Drop rows in columns {} by conditions',
        }
        super_kwargs.update(**kwargs)
        super_kwargs['none_columns'] = 'all'
        super().__init__(**super_kwargs)

    def _transformation(self, df, verbose, fit):
        before_count = len(df)
        columns = self._get_columns(df, fit=fit)
        subdf = df[columns]
        drop_index = ~subdf.apply(self._row_cond, axis=1)
        inter_df = df[drop_index]
        if verbose:
            print(f"{before_count - len(inter_df)} rows dropped.")
        return inter_df


class Schematize(PdPipelineStage):
    """Enforces a column schema on input dataframes.

    Parameters
    ----------
    columns: sequence of labels
        The dataframe schema to enfore on input dataframes.

    Example
    -------
        >>> import pandas as pd; import pdpipe as pdp;
        >>> df = pd.DataFrame([[2, 4, 8],[3, 6, 9]], [1, 2], ['a', 'b', 'c'])
        >>> pdp.Schematize(['a', 'c']).apply(df)
           a  c
        1  2  8
        2  3  9
        >>> pdp.Schematize(['c', 'b']).apply(df)
           c  b
        1  8  4
        2  9  6
    """

    def __init__(self, columns, **kwargs):
        self._columns = _interpret_columns_param(columns)
        self._columns_str = _list_str(self._columns)
        desc = (
            f"Transform input dataframes to the following schema: "
            f"{self._columns_str}"
        )
        exmsg = (
            f"Not all required columns {self._columns_str} "
            f"found in input dataframe!"
        )
        super_kwargs = {
            'exmsg': exmsg,
            'desc': desc,
        }
        super_kwargs.update(**kwargs)
        super().__init__(**super_kwargs)

    def _prec(self, df):
        return set(self._columns).issubset(df.columns)

    def _transform(self, df, verbose=None):
        return df[self._columns]


class DropDuplicates(ColumnsBasedPipelineStage):
    """Drop duplicates in the given columns.

    Parameters
    ----------
    columns: column label or sequence of labels, optional
        The labels of the columns to consider for duplication drop. If not
        populated, duplicates are dropped from all columns.
    exclude_columns : object, iterable or callable, optional
        The label, or an iterable of labels, of columns to exclude, given the
        `columns` parameter. Alternatively, this parameter can be assigned a
        callable returning a labels iterable from an input pandas.DataFrame.
        See pdpipe.cq. Optional. By default no columns are excluded.

    Examples
    --------
        >>> import pandas as pd; import pdpipe as pdp;
        >>> df = pd.DataFrame([[8, 1],[8, 2], [9, 2]], [1,2,3], ['a', 'b'])
        >>> pdp.DropDuplicates('a').apply(df)
           a  b
        1  8  1
        3  9  2
    """

    def __init__(self, columns=None, **kwargs):
        super_kwargs = {
            'columns': columns,
            'desc_temp': 'Drop duplicates in columns {}',
        }
        super_kwargs.update(**kwargs)
        super_kwargs['none_columns'] = 'all'
        super().__init__(**super_kwargs)

    def _transformation(self, df, verbose, fit):
        columns = self._get_columns(df, fit=fit)
        inter_df = df.drop_duplicates(subset=columns)
        if verbose:
            print(f"{len(df) - len(inter_df)} rows dropped.")
        return inter_df


class ColumnDtypeEnforcer(PdPipelineStage):
    """A pipeline stage enforcing column dtypes.

    Parameters
    ----------
    column_to_dtype: dict of column name -> data type
        Use {col: dtype, …}, where col is a column label and dtype is a
        numpy.dtype or Python type to cast one or more of the DataFrame’s
        columns to column-specific types.
    errors: {‘raise’, ‘ignore’}, default ‘raise’
        Control raising of exceptions on invalid data for provided dtype.
        - raise : allow exceptions to be raised
        - ignore : suppress exceptions. On error return original object.

    Example
    -------
        >>> import pandas as pd; import pdpipe as pdp;
        >>> df = pd.DataFrame([[8,'a'],[5,'b']], [1,2], ['num', 'initial'])
        >>> pdp.ColumnDtypeEnforcer({'num': float}).apply(df)
           num initial
        1  8.0       a
        2  5.0       b
    """

    _DEF_COL_DTYPE_ENF_EXC_MSG = (
        "ColumnDtypeEnforcer stage failed because not all columns"
        " {} were found in input dataframe.")

    def __init__(
        self,
        column_to_dtype: Dict[str, object],
        errors: Optional[str] = 'raise',
        **kwargs: object,
    ) -> None:
        self._column_to_dtype = column_to_dtype
        self._errors = errors
        columns_str = _list_str(list(column_to_dtype.keys()))
        suffix = 's' if len(column_to_dtype) > 1 else ''
        keys_set = set(column_to_dtype.keys())

        def _tprec(df: pandas.DataFrame) -> bool:
            return keys_set.issubset(df.columns)
        self._tprec = _tprec
        super_kwargs = {
            'exmsg': ColumnDtypeEnforcer._DEF_COL_DTYPE_ENF_EXC_MSG.format(
                columns_str),
            'desc': f"Renam column{suffix} with {column_to_dtype}",
        }
        super_kwargs.update(**kwargs)
        super().__init__(**super_kwargs)

    def _prec(self, df: pandas.DataFrame) -> bool:
        return self._tprec(df)

    def _transform(
        self,
        df: pandas.DataFrame,
        verbose: bool,
    ) -> pandas.DataFrame:
        return df.astype(
            dtype=self._column_to_dtype,
            copy=True,
            errors=self._errors,
        )



"""Test the TfidfVectorizeTokenLists pipeline stage."""

import pytest
import pandas as pd
import pdpipe as pdp


DF = pd.DataFrame(
    data=[
        [23, ['live', 'full', 'cats', 'mango']],
        [80, ['hovercraft', 'full', 'eels']],
    ],
    columns=['Age', 'Quote'],
)


DF2 = pd.DataFrame(
    data=[
        [29, ['eels', 'full', 'banana']],
    ],
    columns=['Age', 'Quote'],
)


@pytest.mark.parametrize("drop", [True, False])
def test_tfidf_vec(drop):
    tf = pdp.TfidfVectorizeTokenLists('Quote', drop=drop)
    res_df = tf(DF)
    assert 'Age' in res_df.columns
    words = []
    for i, row in DF.iterrows():
        words.extend(row['Quote'])
    uwords = set(words)
    for w in uwords:
        assert w in res_df.columns
    non_zeros = (
        res_df.drop('Quote', axis=1, errors='ignore') > 0
    ).T.sum().values
    for i, row in DF.iterrows():
        assert len(row['Quote']) == non_zeros[i] - 1
    if not drop:
        assert 'Quote' in res_df.columns

    res_df2 = tf(DF2)
    assert 'Age' in res_df2.columns
    if not drop:
        assert 'Quote' in res_df2.columns
    non_zeros2 = (
        res_df2.drop('Quote', axis=1, errors='ignore') > 0
    ).T.sum().values
    assert non_zeros2[0] - 1 == 2


def test_tfidf_vec_hierarchical_labels():
    tf = pdp.TfidfVectorizeTokenLists('Quote', hierarchical_labels=True)
    res_df = tf(DF)
    assert 'Age' in res_df.columns
    words = []
    for i, row in DF.iterrows():
        words.extend(row['Quote'])
    uwords = set(words)
    for w in uwords:
        assert f'Quote_{w}' in res_df.columns
    non_zeros = (
        res_df.drop('Quote', axis=1, errors='ignore') > 0
    ).T.sum().values
    for i, row in DF.iterrows():
        assert len(row['Quote']) == non_zeros[i] - 1



"""Testing columns-based pipeline stages."""

import pandas as pd
import pytest

from pdpipe.core import (
    ColumnsBasedPipelineStage,
    FailedPreconditionError
)


class Drop(ColumnsBasedPipelineStage):
    """A pipeline stage for testing"""

    def __init__(self, columns, errors=None, **kwargs):
        self._errors = errors
        super_kwargs = {
            'columns': columns,
            'desc_temp': 'Drop columns {}',
        }
        super_kwargs.update(**kwargs)
        super().__init__(**super_kwargs)

    def _transformation(self, df, verbose, fit):
        return df.drop(
            self._get_columns(df, fit=fit), axis=1, errors=self._errors)


def _df1():
    return pd.DataFrame(
        data=[[1, 'a', 8], [2, 'b', 9]],
        index=[1, 2],
        columns=['num', 'char', 4]
    )


def test_columns_based_stage():
    df1 = _df1()
    stage = Drop('num')
    res = stage(df1)
    assert 'num' not in res.columns
    assert 'char' in res.columns
    assert 4 in res.columns

    stage = Drop(4)
    res = stage(df1)
    assert 'num' in res.columns
    assert 'char' in res.columns
    assert 4 not in res.columns

    stage = Drop([4])
    res = stage(df1)
    assert 'num' in res.columns
    assert 'char' in res.columns
    assert 4 not in res.columns

    stage = Drop(['num'])
    res = stage(df1)
    assert 'num' not in res.columns
    assert 'char' in res.columns
    assert 4 in res.columns

    def _safe_start_with(string):
        try:
            return string.startswith('n')
        except AttributeError:
            return False
    stage = Drop(lambda df: [
        lbl for lbl in df.columns if _safe_start_with(lbl)
    ])
    res = stage(df1)
    assert 'num' not in res.columns
    assert 'char' in res.columns
    assert 4 in res.columns

    stage = Drop(5)
    with pytest.raises(FailedPreconditionError):
        stage(df1)

    with pytest.raises(ValueError):
        Drop(None)


class Drop2(ColumnsBasedPipelineStage):
    """A pipeline stage for testing"""

    def __init__(self, columns, errors=None, **kwargs):
        self._errors = errors
        super_kwargs = {
            'columns': columns,
            'exmsg': 'Error!',
        }
        super_kwargs.update(**kwargs)
        super().__init__(**super_kwargs)

    def _transformation(self, df, verbose, fit):
        return df.drop(
            self._get_columns(df, fit=fit), axis=1, errors=self._errors)


def test_columns_based_stage2():
    df1 = _df1()
    stage = Drop2('num')
    res = stage(df1)
    assert 'num' not in res.columns
    assert 'char' in res.columns
    assert 4 in res.columns


class Double(ColumnsBasedPipelineStage):
    """A pipeline stage for testing"""

    def __init__(self, columns, errors=None, **kwargs):
        self._errors = errors
        super_kwargs = {
            'columns': columns,
            'desc_temp': 'Drop columns {}',
        }
        super_kwargs.update(**kwargs)
        super_kwargs['none_columns'] = 'all'
        super().__init__(**super_kwargs)

    def _transformation(self, df, verbose, fit):
        inter_df = df
        for col in self._get_columns(df, fit=fit):
            inter_df[col] = inter_df[col] * 2
        return inter_df


def _df_d():
    return pd.DataFrame(
        data=[[1, 2, 3]],
        index=[1],
        columns=['num', 'char', 4]
    )


def test_columns_based_stage_none():
    df = _df_d()
    stage = Double(None)
    res = stage(df)
    assert res.iloc[0, 0] == 2
    assert res.iloc[0, 1] == 4
    assert res.iloc[0, 2] == 6

    stage = Double('num')
    df = _df_d()
    res = stage(df)
    assert res.iloc[0, 0] == 2
    assert res.iloc[0, 1] == 2
    assert res.iloc[0, 2] == 3


class BadNoneColumnsStrArg(ColumnsBasedPipelineStage):
    """A pipeline stage for testing"""

    def __init__(self, columns, errors=None, **kwargs):
        self._errors = errors
        super_kwargs = {
            'columns': columns,
            'desc_temp': 'Drop columns {}',
        }
        super_kwargs.update(**kwargs)
        super_kwargs['none_columns'] = 'badval'
        super().__init__(**super_kwargs)

    def _transformation(self, df, verbose, fit):
        inter_df = df
        for col in self._get_columns(df, fit=fit):
            inter_df[col] = inter_df[col] * 2
        return inter_df


def test_bad_none_columns_str_arg():
    with pytest.raises(ValueError):
        BadNoneColumnsStrArg(['collbl'])


class BadNoneColumnsArg(ColumnsBasedPipelineStage):
    """A pipeline stage for testing"""

    def __init__(self, columns, errors=None, **kwargs):
        self._errors = errors
        super_kwargs = {
            'columns': columns,
            'desc_temp': 'Drop columns {}',
        }
        super_kwargs.update(**kwargs)
        super_kwargs['none_columns'] = FailedPreconditionError()
        super().__init__(**super_kwargs)

    def _transformation(self, df, verbose, fit):
        inter_df = df
        for col in self._get_columns(df, fit=fit):
            inter_df[col] = inter_df[col] * 2
        return inter_df


def test_bad_none_columns_arg():
    with pytest.raises(ValueError):
        BadNoneColumnsArg(['collbl'])


class NumDefaultDrop(ColumnsBasedPipelineStage):
    """A pipeline stage for testing"""

    def __init__(self, columns, errors=None, **kwargs):
        self._errors = errors
        super_kwargs = {
            'columns': columns,
            'desc_temp': 'Drop columns {}',
        }
        super_kwargs.update(**kwargs)
        super_kwargs['none_columns'] = ['num']
        super().__init__(**super_kwargs)

    def _transformation(self, df, verbose, fit):
        keep_cols = [
            x for x in df.columns
            if x not in self._get_columns(df, fit=fit)
        ]
        return df[keep_cols]


def test_columns_based_stage_none_columns_is_list():
    df = _df_d()
    stage = NumDefaultDrop('char')
    res = stage(df)
    assert 'num' in res.columns
    assert 'char' not in res.columns
    assert 4 in res.columns

    stage = NumDefaultDrop(None)
    res = stage(df)
    assert 'num' not in res.columns
    assert 'char' in res.columns
    assert 4 in res.columns



"""Test DropRareTokens stages."""

import pytest
import pandas as pd

from pdpipe.nltk_stages import DropRareTokens


def _some_df():
    return pd.DataFrame(
        data=[[7, ['a', 'a', 'b']], [3, ['b', 'c', 'd']]],
        index=[1, 2],
        columns=["num", "chars"]
    )


def _some_df2():
    return pd.DataFrame(
        data=[[7, ['a', 'c', 'c']], [3, ['b', 'd', 'd']]],
        index=[1, 2],
        columns=["num", "chars"]
    )


@pytest.mark.first
def test_drop_rare():
    df = _some_df()
    drop_rare_stage = DropRareTokens('chars', 1)
    res_df = drop_rare_stage(df)
    assert 'chars' in res_df.columns
    assert res_df['chars'][1] == ['a', 'a', 'b']
    assert res_df['chars'][2] == ['b']

    # see only transform (no fit) when already fitted
    df2 = _some_df2()
    res_df2 = drop_rare_stage(df2)
    assert 'chars' in res_df2.columns
    assert res_df2['chars'][1] == ['a']
    assert res_df2['chars'][2] == ['b']

    # check fit_transform when already fitted
    df2 = _some_df2()
    res_df2 = drop_rare_stage.fit_transform(df2)
    assert 'chars' in res_df2.columns
    assert res_df2['chars'][1] == ['c', 'c']
    assert res_df2['chars'][2] == ['d', 'd']


@pytest.mark.first
def test_drop_rare_w_drop():
    df = _some_df()
    drop_rare_stage = DropRareTokens('chars', 1, drop=False)
    res_df = drop_rare_stage(df, verbose=True)
    assert 'chars' in res_df.columns
    assert 'chars_norare' in res_df.columns
    assert res_df['chars_norare'][1] == ['a', 'a', 'b']
    assert res_df['chars_norare'][2] == ['b']
    assert res_df['chars'][2] == ['b', 'c', 'd']

    # see only transform (no fit) when already fitted
    df2 = _some_df2()
    res_df2 = drop_rare_stage(df2, verbose=True)
    assert 'chars' in res_df2.columns
    assert 'chars_norare' in res_df2.columns
    assert res_df2['chars_norare'][1] == ['a']
    assert res_df2['chars_norare'][2] == ['b']
    assert res_df2['chars'][1] == ['a', 'c', 'c']
    assert res_df2['chars'][2] == ['b', 'd', 'd']

    # check fit_transform when already fitted
    df2 = _some_df2()
    res_df2 = drop_rare_stage.fit_transform(df2)
    assert 'chars' in res_df2.columns
    assert 'chars_norare' in res_df2.columns
    assert res_df2['chars_norare'][1] == ['c', 'c']
    assert res_df2['chars_norare'][2] == ['d', 'd']
    assert res_df2['chars'][1] == ['a', 'c', 'c']
    assert res_df2['chars'][2] == ['b', 'd', 'd']



"""Test the DropTokensByList pipeline stage."""

import pandas as pd
import pdpipe as pdp


def test_drop_tokens_by_list_short():
    data = [[4, ["a", "bad", "cat"]], [5, ["bad", "not", "good"]]]
    df = pd.DataFrame(data, [1, 2], ["age", "text"])
    filter_tokens = pdp.DropTokensByList('text', ['bad'])
    res_df = filter_tokens(df)
    assert 'age' in res_df.columns
    assert 'text' in res_df.columns
    assert 'bad' not in res_df.loc[1]['text']
    assert 'a' in res_df.loc[1]['text']
    assert 'cat' in res_df.loc[1]['text']
    assert 'bad' not in res_df.loc[2]['text']
    assert 'not' in res_df.loc[2]['text']
    assert 'good' in res_df.loc[2]['text']


def test_drop_tokens_by_list_short_no_drop():
    data = [[4, ["a", "bad", "cat"]], [5, ["bad", "not", "good"]]]
    df = pd.DataFrame(data, [1, 2], ["age", "text"])
    filter_tokens = pdp.DropTokensByList('text', ['bad'], drop=False)
    res_df = filter_tokens(df)
    assert 'age' in res_df.columns
    assert 'text' in res_df.columns
    assert 'text_filtered' in res_df.columns
    assert 'bad' not in res_df.loc[1]['text_filtered']
    assert 'a' in res_df.loc[1]['text_filtered']
    assert 'cat' in res_df.loc[1]['text_filtered']
    assert 'bad' not in res_df.loc[2]['text_filtered']
    assert 'not' in res_df.loc[2]['text_filtered']
    assert 'good' in res_df.loc[2]['text_filtered']


def test_drop_tokens_by_long_short():
    data = [[4, ["a", "bad", "cat"]], [5, ["bad", "not", "good"]]]
    df = pd.DataFrame(data, [1, 2], ["age", "text"])
    bad = ['z', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'bad']
    filter_tokens = pdp.DropTokensByList('text', bad)
    res_df = filter_tokens(df)
    assert 'age' in res_df.columns
    assert 'text' in res_df.columns
    assert 'bad' not in res_df.loc[1]['text']
    assert 'a' in res_df.loc[1]['text']
    assert 'cat' in res_df.loc[1]['text']
    assert 'bad' not in res_df.loc[2]['text']
    assert 'not' in res_df.loc[2]['text']
    assert 'good' in res_df.loc[2]['text']



"""Text processing pdpipe pipeline stages."""

import re

from pdpipe.col_generation import ApplyByCols


class RegexReplace(ApplyByCols):
    """A pipeline stage replacing regex occurences in a text column.

    Parameters
    ----------
    columns : str or list-like
        Names of columns on which to apply regex replacement.
    pattern : str
        The regex whose occurences will be replaced.
    replace : str
        The replacement string to use. This is equivalent to repl in re.sub.
    result_columns : str or list-like, default None
        The names of the new columns resulting from the mapping operation. Must
        be of the same length as columns. If None, behavior depends on the
        drop parameter: If drop is True, the name of the source column is used;
        otherwise, the name of the source column is used with the suffix
        '_reg'.
    drop : bool, default True
        If set to True, source columns are dropped after being transformed.

    Example
    -------
        >>> import pandas as pd; import pdpipe as pdp;
        >>> data = [[4, "more than 12"], [5, "with 5 more"]]
        >>> df = pd.DataFrame(data, [1,2], ["age","text"])
        >>> clean_num = pdp.RegexReplace('text', r'\\b[0-9]+\\b', "NUM")
        >>> clean_num(df)
           age           text
        1    4  more than NUM
        2    5  with NUM more
    """  # noqa: W605

    class RegexReplacer(object):
        """A pickle-able regex replacement function."""

        def __init__(self, pattern_obj, replace_text):
            self.pattern_obj = pattern_obj
            self.replace_text = replace_text

        def __call__(self, x):
            return self.pattern_obj.sub(self.replace_text, x)

    def __init__(
        self,
        columns,
        pattern,
        replace,
        result_columns=None,
        drop=True,
        func_desc=None,
        **kwargs
    ):
        self._pattern = pattern
        self._replace = replace
        self._pattern_obj = re.compile(pattern)
        desc_temp = "Replacing appearances of {} with '{}' in column {{}}"
        desc_temp = desc_temp.format(pattern, replace)
        super_kwargs = {
            'columns': columns,
            'func': RegexReplace.RegexReplacer(
                self._pattern_obj, self._replace),
            'suffix': '_regex',
            'result_columns': result_columns,
            'drop': drop,
            'desc_temp': desc_temp,
        }
        super_kwargs.update(**kwargs)
        super().__init__(**super_kwargs)


class DropTokensByLength(ApplyByCols):
    """A pipeline stage removing tokens by length in string-token list columns.

    Parameters
    ----------
    columns : str or list-like
        Names of token list columns on which to apply token filtering.
    min_len : int
        The minimum length of tokens to keep. Tokens of shorter length are
        removed from all token lists.
    max_len : int, default None
        The maximum length of tokens to keep. If provided, tokens of longer
        length are removed from all token lists.
    result_columns : str or list-like, default None
        The names of the new columns resulting from the mapping operation.
        Must be of the same length as columns. If None, behavior depends on
        the drop parameter: If drop is True, the name of the source column
        is used; otherwise, the name of the source column is used with the
        suffix '_filtered'.
    drop : bool, default True
        If set to True, source columns are dropped after being transformed.

    Example
    -------
        >>> import pandas as pd; import pdpipe as pdp;
        >>> data = [[4, ["a", "bad", "nice"]], [5, ["good", "university"]]]
        >>> df = pd.DataFrame(data, [1,2], ["age","text"])
        >>> filter_tokens = pdp.DropTokensByLength('text', 3, 5)
        >>> filter_tokens(df)
           age         text
        1    4  [bad, nice]
        2    5       [good]
    """  # noqa: W605

    class MinLengthTokenFilter(object):

        def __init__(self, min_len):
            self.min_len = min_len

        def __call__(self, token_list):
            return [x for x in token_list if len(x) >= self.min_len]

    class MinMaxLengthTokenFilter(object):

        def __init__(self, min_len, max_len):
            self.min_len = min_len
            self.max_len = max_len

        def __call__(self, token_list):
            return [
                x for x in token_list
                if len(x) >= self.min_len and len(x) <= self.max_len
            ]

    def __init__(
        self, columns, min_len, max_len=None, result_columns=None, drop=True,
        **kwargs
    ):
        self._min_len = min_len
        self._max_len = max_len
        token_filter = DropTokensByLength.MinLengthTokenFilter(min_len)
        cond_str = f" > {min_len}"
        if max_len:
            token_filter = DropTokensByLength.MinMaxLengthTokenFilter(
                min_len=min_len, max_len=max_len)
            cond_str += f" < {max_len}"
        desc_temp = "Filtering out tokens of length{} in columns {{}}"
        desc_temp = desc_temp.format(cond_str)
        super_kwargs = {
            'columns': columns,
            'func': token_filter,
            'result_columns': result_columns,
            'drop': drop,
            'suffix': "_filtered",
            'desc_temp': desc_temp,
        }
        super_kwargs.update(**kwargs)
        super().__init__(**super_kwargs)


class DropTokensByList(ApplyByCols):
    """A pipeline stage removing specific tokens in string-token list columns.

    Parameters
    ----------
    columns : str or list-like
        Names of token list columns on which to apply token filtering.
    bad_tokens : list of str
        The list of string tokens to remove from all token lists.
    result_columns : str or list-like, default None
        The names of the new columns resulting from the mapping operation.
        Must be of the same length as columns. If None, behavior depends on
        the drop parameter: If drop is True, the name of the source column
        is used; otherwise, the name of the source column is used with the
        suffix '_filtered'.
    drop : bool, default True
        If set to True, source columns are dropped after being transformed.

    Example
    -------
        >>> import pandas as pd; import pdpipe as pdp;
        >>> data = [[4, ["a", "bad", "cat"]], [5, ["bad", "not", "good"]]]
        >>> df = pd.DataFrame(data, [1,2], ["age","text"])
        >>> filter_tokens = pdp.DropTokensByList('text', ['bad'])
        >>> filter_tokens(df)
           age         text
        1    4     [a, cat]
        2    5  [not, good]
    """  # noqa: W605

    class ListTokenFilter(object):

        def __init__(self, bad_tokens):
            self.bad_tokens = bad_tokens

        def __call__(self, token_list):
            return [x for x in token_list if x not in self.bad_tokens]

    def __init__(
        self, columns, bad_tokens, result_columns=None, drop=True,
        **kwargs
    ):
        self._bad_tokens = bad_tokens
        cond_str = ""
        if len(bad_tokens) < 10:
            cond_str = " in list [" + " ".join(bad_tokens) + "]"
        base_str = "Filtering out tokens{} in columns {{}}"
        desc_temp = base_str.format(cond_str)
        super_kwargs = {
            'columns': columns,
            'func': DropTokensByList.ListTokenFilter(bad_tokens),
            'result_columns': result_columns,
            'drop': drop,
            'suffix': "_filtered",
            'desc_temp': desc_temp,
        }
        super_kwargs.update(**kwargs)
        super().__init__(**super_kwargs)



"""Testing the ColReorder stage."""

import pytest
import pandas as pd
import pdpipe as pdp

from pdpipe import Schematize
from pdpipe.exceptions import FailedPreconditionError


def _df():
    return pd.DataFrame([[2, 4, 8], [3, 6, 9]], [1, 2], ['a', 'b', 'c'])


def _df2():
    return pd.DataFrame([[2, 4], [3, 6]], [1, 2], ['a', 'b'])


def _df3():
    return pd.DataFrame([[2, 4], [3, 6]], [1, 2], ['a', 'c'])


def _df4():
    return pd.DataFrame([[2, 4, 8], [3, 6, 9]], [1, 2], ['b', 'c', 'a'])


def test_schematize():
    df = _df()
    stage = Schematize(['a', 'c'])
    res = stage(df)
    assert list(res.columns) == ['a', 'c']
    assert res.iloc[0, 0] == 2
    assert res.iloc[1, 0] == 3
    assert res.iloc[0, 1] == 8
    assert res.iloc[1, 1] == 9

    stage = Schematize(['c', 'b'])
    res = stage(df)
    assert list(res.columns) == ['c', 'b']
    assert res.iloc[0, 0] == 8
    assert res.iloc[1, 0] == 9
    assert res.iloc[0, 1] == 4
    assert res.iloc[1, 1] == 6

    stage = Schematize(['a', 'g'])
    with pytest.raises(FailedPreconditionError):
        stage(df)


@pytest.mark.xfail
def test_schematize_with_cq():
    # fit the stage on an [a, b] df
    df = _df2()
    stage = Schematize(pdp.cq.AllColumns())
    res = stage(df)
    assert list(res.columns) == ['a', 'b']

    # pass an [a, b, c] df to the fitted stage, make sure c is dropped
    df = _df()
    res = stage(df)
    assert list(res.columns) == ['a', 'b']

    # pass an [b, c, a] df to the fitted stage, make sure c is dropped and
    # a, b are correctly ordered
    df = _df4()
    res = stage(df)
    assert list(res.columns) == ['a', 'b']

    # check that a df with [a, c] columns fails a stage fitted with [a, b]
    with pytest.raises(FailedPreconditionError):
        stage(_df3())



"""Testing application context.."""

from random import randint

import pandas as pd
import pytest

from pdpipe.core import (
    PdPipelineStage,
    AdHocStage,
    PdPipeline
)
from pdpipe.util import out_of_place_col_insert
from pdpipe.core import PdpApplicationContext


def _test_df():
    return pd.DataFrame(
        data=[[1, 2, 'a'], [2, 4, 'b']],
        index=[1, 2],
        columns=['num1', 'num2', 'char']
    )


class PutContextStage(PdPipelineStage):
    """A pipeline stage for testing"""

    def __init__(self, colname, **kwargs):
        self.colname = colname
        self.randi = randint(1, 832)
        super().__init__(exraise=False, **kwargs)

    def _prec(self, df):
        return self.colname in df.columns

    def _transform(self, df, verbose):
        self.fit_context['a'] = self.randi
        return df.drop([self.colname], axis=1)


class UseContextStage(PdPipelineStage):
    """A pipeline stage for testing"""

    def __init__(self, colname, **kwargs):
        self.colname = colname
        super().__init__(exraise=False, **kwargs)

    def _prec(self, df):
        return self.colname in df.columns

    def _transform(self, df, verbose):
        val = self.fit_context['a']
        source_col = df[self.colname]
        loc = df.columns.get_loc(self.colname) + 1
        series = source_col.apply(lambda x: x + val)
        inter_df = out_of_place_col_insert(
            df=df,
            series=series,
            loc=loc,
            column_name=self.colname + "+val",
        )
        return inter_df


def test_application_context():
    """Testing something."""
    put_context = PutContextStage('char')
    val = put_context.randi
    use_context = UseContextStage('num1')
    pipeline = PdPipeline([put_context, use_context])
    assert len(pipeline) == 2
    df = _test_df()
    res_df = pipeline.apply(df, verbose=True)
    assert 'num1' in res_df.columns
    assert 'num1+val' in res_df.columns
    assert 'num2' in res_df.columns
    assert 'char' not in res_df.columns
    num1 = df['num1']
    resnum = res_df['num1+val']
    assert resnum.iloc[0] == val + num1.iloc[0]
    assert resnum.iloc[1] == val + num1.iloc[1]

    val2 = pipeline.fit_context['a']
    assert val2 == val
    # check locking works
    pipeline.fit_context['a'] = 0
    val2 = pipeline.fit_context['a']
    assert val2 == val
    del pipeline.fit_context['a']
    val2 = pipeline.fit_context['a']
    assert val2 == val
    pipeline.fit_context.pop('a', 0)
    val2 = pipeline.fit_context['a']
    assert val2 == val
    pipeline.fit_context.clear()
    val2 = pipeline.fit_context['a']
    assert val2 == val


def test_application_context_unit():
    context = PdpApplicationContext()
    context['a'] = 1
    assert context['a'] == 1
    assert context.pop('a', 0) == 1
    with pytest.raises(KeyError):
        context['a']
    context['a'] = 1
    del context['a']
    with pytest.raises(KeyError):
        context['a']
    context['a'] = 1
    context.clear()
    with pytest.raises(KeyError):
        context['a']
    context.update({'b': 2})
    assert context['b'] == 2
    outer_context = PdpApplicationContext(context)
    assert outer_context.fit_context() == context

    # now check locking
    context['a'] = 1
    context.lock()
    assert context.pop('a', 0) == 1
    assert context['a'] == 1
    del context['a']
    assert context['a'] == 1
    context.clear()
    assert context['a'] == 1
    context.update({'c': 3})
    with pytest.raises(KeyError):
        context['c']


CONTEXT_VAR = 'a'
CONTEXT_NUM = 5
SRC_LBL = 'num1'
TRGT_LBL = 'res'


def put_context(df, fit_context, verbose):
    if verbose:
        print('Verbose in put_context test!')
    fit_context[CONTEXT_VAR] = CONTEXT_NUM
    return df


def use_context(df, fit_context):
    val = fit_context[CONTEXT_VAR]
    source_col = df[SRC_LBL]
    loc = df.columns.get_loc(SRC_LBL) + 1
    series = source_col.apply(lambda x: x + val)
    inter_df = out_of_place_col_insert(
        df=df,
        series=series,
        loc=loc,
        column_name=TRGT_LBL,
    )
    return inter_df


def test_context_with_adhoc_stage():
    pipeline = PdPipeline([
        AdHocStage(transform=put_context),
        AdHocStage(transform=use_context),
    ])
    assert len(pipeline) == 2
    df = _test_df()
    res_df = pipeline.apply(df, verbose=True)
    assert 'num1' in res_df.columns
    assert TRGT_LBL in res_df.columns
    assert 'num2' in res_df.columns
    assert 'char' in res_df.columns
    num1 = df[SRC_LBL]
    resnum = res_df[TRGT_LBL]
    assert resnum.iloc[0] == CONTEXT_NUM + num1.iloc[0]
    assert resnum.iloc[1] == CONTEXT_NUM + num1.iloc[1]



"""Testing ColRename pipeline stages."""

import pandas as pd

from pdpipe.basic_stages import SetIndex


def test_set_index():
    """Testing the SetIndex pipeline stage."""
    df = pd.DataFrame(
        data=[[8, 'a', 4], [5, 'b', 92]],
        index=[1, 2],
        columns=['num', 'char', 'w']
    )
    res_df = SetIndex('num', drop=True).apply(df)
    assert 'w' in res_df.columns
    assert 'num' not in res_df.columns
    assert 'char' in res_df.columns
    assert res_df.index.name == 'num'

    res_df = SetIndex(['w', 'num'], drop=True).apply(df)
    assert 'w' not in res_df.columns
    assert 'num' not in res_df.columns
    assert 'char' in res_df.columns



"""Testing basic pipeline stages."""
from builtins import ValueError

import pandas as pd
import pytest

from pdpipe.core import (
    PdPipelineStage,
    PdPipeline
)
from pdpipe import make_pdpipeline, ColByFrameFunc, ColDrop
from pdpipe.exceptions import PipelineApplicationError


def _test_df():
    return pd.DataFrame(
        data=[[1, 2, 'a'], [2, 4, 'b']],
        index=[1, 2],
        columns=['num1', 'num2', 'char']
    )


class SilentDropStage(PdPipelineStage):
    """A pipeline stage for testing"""

    def __init__(self, colname, **kwargs):
        self.colname = colname
        super().__init__(exraise=False, **kwargs)

    def _prec(self, df):
        return self.colname in df.columns

    def _transform(self, df, verbose):
        return df.drop([self.colname], axis=1)


@pytest.mark.parametrize("time", [True, False])
def test_two_stage_pipeline_stage(time):
    """Testing something."""
    drop_num1 = SilentDropStage('num1')
    drop_num2 = SilentDropStage('num2')
    pipeline = PdPipeline([drop_num1, drop_num2])
    assert len(pipeline) == 2
    df = _test_df()
    res_df = pipeline.apply(df, verbose=True)
    assert 'num1' not in res_df.columns
    assert 'num2' not in res_df.columns
    assert 'char' in res_df.columns
    str(pipeline)
    pipeline.memory_report()

    # test fit
    df = _test_df()
    res_df = pipeline.fit(df, verbose=True, time=time)
    for x in ['num1', 'num2', 'char']:
        assert x in res_df.columns

    # test transform
    df = _test_df()
    res_df = pipeline.transform(df, verbose=True, time=time)
    assert 'num1' not in res_df.columns
    assert 'num2' not in res_df.columns
    assert 'char' in res_df.columns

    # test fit_transform
    df = _test_df()
    res_df = pipeline.fit_transform(df, verbose=True, time=time)

    # test get_transformer
    trs = lambda pipline: pipeline[:1]  # noqa: E731
    pipeline = PdPipeline([drop_num1, drop_num2], transformer_getter=trs)
    transformer = pipeline.get_transformer()
    res_df = transformer(df, verbose=True)
    assert 'num1' not in res_df.columns
    assert 'num2' in res_df.columns
    assert 'char' in res_df.columns


def test_make_pdpipeline():
    """Testing something."""
    drop_num1 = SilentDropStage('num1')
    drop_num2 = SilentDropStage('num2')
    pipeline = make_pdpipeline(drop_num1, drop_num2)
    assert len(pipeline) == 2
    df = _test_df()
    res_df = pipeline.apply(df, verbose=True)
    assert 'num1' not in res_df.columns
    assert 'num2' not in res_df.columns
    assert 'char' in res_df.columns


def test_pipeline_stage_addition():
    """Testing something."""
    drop_num1 = SilentDropStage('num1')
    drop_num2 = SilentDropStage('num2')
    pipeline = drop_num1 + drop_num2
    assert len(pipeline) == 2
    df = _test_df()
    res_df = pipeline.apply(df, verbose=True)
    assert 'num1' not in res_df.columns
    assert 'num2' not in res_df.columns
    assert 'char' in res_df.columns


def test_pipeline_to_pipeline_stage_addition():
    """Testing something."""
    drop_num1 = SilentDropStage('num1')
    drop_num2 = SilentDropStage('num2')
    pipeline = PdPipeline([drop_num1])
    assert len(pipeline) == 1
    pipeline = pipeline + drop_num2
    assert len(pipeline) == 2
    df = _test_df()
    res_df = pipeline.apply(df, verbose=True)
    assert 'num1' not in res_df.columns
    assert 'num2' not in res_df.columns
    assert 'char' in res_df.columns


def test_pipeline_stage_to_pipeline_addition():
    """Testing something."""
    drop_num1 = SilentDropStage('num1')
    drop_num2 = SilentDropStage('num2')
    pipeline = PdPipeline([drop_num1])
    assert len(pipeline) == 1
    pipeline = drop_num2 + pipeline
    assert len(pipeline) == 2
    df = _test_df()
    res_df = pipeline.apply(df, verbose=True)
    assert 'num1' not in res_df.columns
    assert 'num2' not in res_df.columns
    assert 'char' in res_df.columns


def test_pipeline_to_pipeline_addition():
    """Testing something."""
    drop_num1 = SilentDropStage('num1')
    drop_num2 = SilentDropStage('num2')
    pipeline1 = PdPipeline([drop_num1])
    pipeline2 = PdPipeline([drop_num2])
    pipeline = pipeline1 + pipeline2
    assert len(pipeline) == 2
    assert pipeline[0] == drop_num1
    assert pipeline[1] == drop_num2
    df = _test_df()
    res_df = pipeline.apply(df, verbose=True)
    assert 'num1' not in res_df.columns
    assert 'num2' not in res_df.columns
    assert 'char' in res_df.columns


def test_pipeline_to_int_addition():
    """Testing something."""
    pipeline = PdPipeline([SilentDropStage('num1')])
    with pytest.raises(TypeError):
        res = pipeline + 43
        assert not isinstance(res, PdPipeline)


def test_pipeline_index():
    """Testing something."""
    df = _test_df()
    drop_num1 = SilentDropStage('num1')
    drop_num2 = SilentDropStage('num2')
    drop_char = SilentDropStage('char')
    pipeline = PdPipeline([drop_num1, drop_num2, drop_char])
    assert len(pipeline) == 3
    assert pipeline[0] == drop_num1
    assert 'num1' not in pipeline[0](df).columns
    assert pipeline[1] == drop_num2
    assert 'num2' not in pipeline[1](df).columns
    assert pipeline[2] == drop_char
    assert 'char' not in pipeline[2](df).columns


def test_pipeline_slice():
    """Testing something."""
    drop_num1 = SilentDropStage('num1')
    drop_num2 = SilentDropStage('num2')
    drop_char = SilentDropStage('char')
    pipeline = PdPipeline([drop_num1, drop_num2, drop_char])
    assert len(pipeline) == 3
    pipeline = pipeline[0:2]
    assert len(pipeline) == 2
    df = _test_df()
    res_df = pipeline.apply(df, verbose=True)
    assert 'num1' not in res_df.columns
    assert 'num2' not in res_df.columns
    assert 'char' in res_df.columns


def test_pipeline_slice_by_name():
    """Testing something."""
    drop_num1 = SilentDropStage('num1', name='dropNum1')
    drop_num2 = SilentDropStage('num2', name='dropNum2')
    drop_char = SilentDropStage('char', name='dropChar')
    pipeline = PdPipeline([drop_num1, drop_num2, drop_char])
    assert len(pipeline) == 3
    pipeline = pipeline[['dropNum1', 'dropNum2']]
    assert len(pipeline) == 2
    assert pipeline['dropNum1'] == drop_num1
    with pytest.raises(ValueError) as e:
        pipeline['dropChar']
    assert str(e.value) == "'dropChar' is not exist."
    df = _test_df()
    res_df = pipeline.apply(df, verbose=True)
    assert 'num1' not in res_df.columns
    assert 'num2' not in res_df.columns
    assert 'char' in res_df.columns


@pytest.mark.parametrize("time", [True, False])
def test_pipeline_error(time):
    """Test exceptions at pipeline level"""

    # test fit
    df = _test_df()

    def _func(df):
        return df['num1'] == df['num3']
    pipeline = PdPipeline([ColByFrameFunc("Equality", _func), ColDrop("B")])
    with pytest.raises(PipelineApplicationError):
        pipeline.fit(df, verbose=True, time=time)

    # test transform
    df = _test_df()
    with pytest.raises(PipelineApplicationError):
        pipeline.transform(df, verbose=True, time=time)

    # test fit_transform
    df = _test_df()
    with pytest.raises(PipelineApplicationError):
        pipeline.fit_transform(df, verbose=True, time=time)



"""Testing ApplyToRows pipeline stages."""

import pandas as pd

from pdpipe.col_generation import ApplyToRows


def _some_df():
    return pd.DataFrame(
        data=[[3, 2143], [10, 1321], [7, 1255]],
        index=[1, 2, 3],
        columns=["years", "avg_revenue"]
    )


def _total_rev(row):
    return row['years'] * row['avg_revenue']


def test_applytorows():
    """Testing ApplyToRows pipeline stages."""
    df = _some_df()
    cbf_stage = ApplyToRows(_total_rev, 'total_revenue')
    res_df = cbf_stage(df)
    assert res_df.columns.get_loc('total_revenue') == 2
    assert res_df['total_revenue'][1] == 3 * 2143
    assert res_df['total_revenue'][2] == 10 * 1321
    assert res_df['total_revenue'][3] == 7 * 1255


def test_applytorows_with_follow_column():
    """Testing ApplyToRows pipeline stages."""
    df = _some_df()
    cbf_stage = ApplyToRows(_total_rev, 'total_revenue', follow_column='years')
    res_df = cbf_stage(df)
    assert res_df.columns.get_loc('total_revenue') == 1
    assert res_df['total_revenue'][1] == 3 * 2143
    assert res_df['total_revenue'][2] == 10 * 1321
    assert res_df['total_revenue'][3] == 7 * 1255


def _num_df():
    return pd.DataFrame(
        data=[[1, 2, 'a'], [2, 4, 'b']],
        index=[1, 2],
        columns=['num1', 'num2', 'char']
    )


def _sum_and_diff(row):
    return pd.Series(
        {'sum': row['num1'] + row['num2'], 'diff': row['num1'] - row['num2']})


def test_applytorows_with_df_generation():
    """Testing ApplyToRows pipeline stages."""
    df = _num_df()
    cbf_stage = ApplyToRows(_sum_and_diff)
    res_df = cbf_stage(df)
    assert 'sum' in res_df.columns
    assert 'diff' in res_df.columns
    assert res_df['sum'][1] == 3
    assert res_df['sum'][2] == 6
    assert res_df['diff'][1] == -1
    assert res_df['diff'][2] == -2


def test_applytorows_with_df_generation_and_optionals():
    """Testing ApplyToRows pipeline stages."""
    df = _num_df()
    cbf_stage = ApplyToRows(
        func=_sum_and_diff,
        func_desc="calculates sum and diff of num1 and num2",
        prec=lambda df: 'num1' in df.columns and 'num2' in df.columns
    )
    res_df = cbf_stage(df)
    assert 'sum' in res_df.columns
    assert 'diff' in res_df.columns
    assert res_df['sum'][1] == 3
    assert res_df['sum'][2] == 6
    assert res_df['diff'][1] == -1
    assert res_df['diff'][2] == -2


def test_applytorows_with_df_generation_follow():
    """Testing ApplyToRows pipeline stages."""
    df = _num_df()
    cbf_stage = ApplyToRows(_sum_and_diff, follow_column='num1')
    res_df = cbf_stage(df)
    print(res_df.columns)
    assert res_df.columns.get_loc('sum') == 2
    assert res_df.columns.get_loc('diff') == 1
    assert 'sum' in res_df.columns
    assert 'diff' in res_df.columns
    assert res_df['sum'][1] == 3
    assert res_df['sum'][2] == 6
    assert res_df['diff'][1] == -1
    assert res_df['diff'][2] == -2



"""Testing basic pipeline stages."""

import pandas as pd

from pdpipe.core import (
    AdHocStage
)


def _test_df():
    return pd.DataFrame(
        data=[[1, 'a'], [2, 'b']],
        index=[1, 2],
        columns=['num', 'char']
    )


def test_adhoc_stage():
    """Testing ad hoc stages."""
    test_stage = AdHocStage(
        transform=lambda df: df.drop(['num'], axis=1),
        prec=lambda df: 'num' in df.columns
    )
    df = _test_df()
    res_df = test_stage.apply(df, verbose=True)
    assert 'num' not in res_df.columns
    assert 'char' in res_df.columns


def test_adhoc_stage_no_prec():
    """Testing ad hoc stages."""
    test_stage = AdHocStage(
        transform=lambda df: df.drop(['num'], axis=1),
        prec=None
    )
    df = _test_df()
    res_df = test_stage.apply(df, verbose=True)
    assert 'num' not in res_df.columns
    assert 'char' in res_df.columns


def test_adhoc_stage_w_fit_transform():
    """Testing ad hoc stages with fit_transform."""
    test_stage = AdHocStage(
        transform=lambda df: df.drop(['num'], axis=1),
        fit_transform=lambda df: df.drop(['char'], axis=1),
    )

    df = _test_df()
    res_df = test_stage.apply(df, verbose=True)
    assert 'num' in res_df.columns
    assert 'char' not in res_df.columns

    df = _test_df()
    res_df = test_stage.apply(df, verbose=True)
    assert 'num' not in res_df.columns
    assert 'char' in res_df.columns



"""Testing ApplyToRows pipeline stages."""

# flake8: noqa: E712

import pytest
import pandas as pd

from pdpipe import ColByFrameFunc
from pdpipe.exceptions import PipelineApplicationError


def _some_df():
    return pd.DataFrame(
        data=[[3, 3], [2, 4], [1, 5]],
        index=[1, 2, 3],
        columns=["A", "B"]
    )


def _are_a_b_equal(df):
    return df['A'] == df['B']


def test_colbyframefunc():
    df = _some_df()
    cbf_stage = ColByFrameFunc('A==B', _are_a_b_equal)
    res_df = cbf_stage(df)
    assert res_df.columns.get_loc('A==B') == 2
    assert res_df['A==B'][1]
    assert not res_df['A==B'][2]
    assert not res_df['A==B'][3]


def test_colbyframefunc_follow():
    df = _some_df()
    cbf_stage = ColByFrameFunc(
        'A==B', _are_a_b_equal, follow_column='A', func_desc='R')
    res_df = cbf_stage(df)
    assert res_df.columns.get_loc('A==B') == 1
    assert res_df['A==B'][1]
    assert not res_df['A==B'][2]
    assert not res_df['A==B'][3]


def _are_a_c_equal(df):
    return df['A'] == df['C']


def test_colbyframefunc_error():
    df = _some_df()
    cbf_stage = ColByFrameFunc('A==B', _are_a_c_equal)
    with pytest.raises(PipelineApplicationError):
        cbf_stage(df)


def test_colbyframefunc_before_column():
    df = _some_df()
    cbf_stage = ColByFrameFunc(
        'A==B', _are_a_b_equal, before_column='B', func_desc='R')
    res_df = cbf_stage(df)
    assert res_df.columns.get_loc('A==B') == 1
    assert res_df['A==B'][1]
    assert not res_df['A==B'][2]
    assert not res_df['A==B'][3]


def test_colbyframefunc_replace():
    df = _some_df()
    cbf_stage = ColByFrameFunc(
        'B', _are_a_b_equal, before_column='B', func_desc='R')
    res_df = cbf_stage(df)
    assert res_df.columns.get_loc('B') == 1
    assert res_df['B'][1]
    assert not res_df['B'][2]
    assert not res_df['B'][3]



"""Setup for the pdpipe package."""

# !/usr/bin/env python
# -*- coding: utf-8 -*-

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

import versioneer


README_RST = ''
with open('README.rst', encoding="utf-8") as f:
    README_RST = f.read()

INSTALL_REQUIRES = [
    'pandas>=0.18.0', 'sortedcontainers', 'tqdm', 'strct', 'skutil>=0.0.15',
]
TEST_REQUIRES = [
    # testing and coverage
    'pytest', 'coverage', 'pytest-cov', 'pytest-ordering',
    # non-testing packagesrequired by tests, not by the package
    'scikit-learn', 'pdutil', 'nltk',
    # to be able to run `python setup.py checkdocs`
    'collective.checkdocs', 'pygments',
]


setup(
    name='pdpipe',
    description="Easy pipelines for pandas.",
    long_description=README_RST,
    long_description_content_type='text/x-rst',
    author="Shay Palachy",
    author_email="shaypal5@gmail.com",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    url='https://pdpipe.github.io/pdpipe/',
    license="MIT",
    packages=['pdpipe'],
    install_requires=INSTALL_REQUIRES,
    extras_require={
        'sklearn': ['scikit-learn'],
        'nltk': ['nltk'],
        'test': TEST_REQUIRES
    },
    platforms=['any'],
    keywords='pandas dataframe pipeline data',
    classifiers=[
        # Trove classifiers
        # (https://pypi.python.org/pypi?%3Aaction=list_classifiers)
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Intended Audience :: Developers',
    ],
)



"""Test the RegexReplace pipeline stage."""

import pandas as pd
import pdpipe as pdp


DF = pd.DataFrame(
    data=[[4, "more than 12"], [5, "with 5 more"]],
    index=[1, 2],
    columns=["age", "text"],
)


def test_regex_replace():
    clean_num = pdp.RegexReplace('text', r'\b[0-9]+\b', "NUM")
    res_df = clean_num(DF)
    assert 'age' in res_df.columns
    assert 'text' in res_df.columns
    assert res_df.loc[1]['text'] == 'more than NUM'
    assert res_df.loc[2]['text'] == 'with NUM more'


def test_regex_replace_no_drop():
    clean_num = pdp.RegexReplace('text', r'\b[0-9]+\b', "NUM", drop=False)
    res_df = clean_num(DF)
    assert 'age' in res_df.columns
    assert 'text' in res_df.columns
    assert 'text_regex' in res_df.columns
    assert res_df.loc[1]['text'] == 'more than 12'
    assert res_df.loc[2]['text'] == 'with 5 more'
    assert res_df.loc[1]['text_regex'] == 'more than NUM'
    assert res_df.loc[2]['text_regex'] == 'with NUM more'



"""Testing the FitOnly wrapper pipeline stage."""

import pandas as pd
import pytest

from pdpipe import ColDrop, FitOnly
from pdpipe.exceptions import FailedPreconditionError


def _test_df():
    return pd.DataFrame(
        data=[[1, 2, 'a'], [2, 4, 'b']],
        index=[1, 2],
        columns=['num1', 'num2', 'char']
    )


def test_fitonly_w_coldrop():
    df = _test_df()
    stage = FitOnly(ColDrop('num1'))
    res = stage(df)
    assert 'num1' not in res.columns
    assert 'num2' in res.columns
    assert 'char' in res.columns

    df = _test_df()
    res = stage(df)
    assert 'num1' in res.columns
    assert 'num2' in res.columns
    assert 'char' in res.columns

    df = _test_df()
    res = stage.fit_transform(df)
    assert 'num1' not in res.columns
    assert 'num2' in res.columns
    assert 'char' in res.columns

    df = _test_df()
    res = stage(df, verbose=True)
    assert 'num1' in res.columns
    assert 'num2' in res.columns
    assert 'char' in res.columns


def test_fitonly_w_coldrop_missing_col():
    df = _test_df()
    stage = FitOnly(ColDrop('num3'))
    with pytest.raises(FailedPreconditionError):
        stage.apply(df)



"""Testing basic pipeline stages."""

import pandas as pd

from pdpipe.cq import StartWith
from pdpipe.basic_stages import ValDrop


def test_valdrop_with_columns():
    """Testing the ColDrop pipeline stage."""
    df = pd.DataFrame([[1, 4], [4, 5], [18, 11]], [1, 2, 3], ['a', 'b'])
    res_df = ValDrop([4], 'a').apply(df)
    assert 1 in res_df.index
    assert 2 not in res_df.index
    assert 3 in res_df.index


def test_valdrop_with_columns_verbose():
    """Testing the ColDrop pipeline stage."""
    df = pd.DataFrame([[1, 4], [4, 5], [18, 11]], [1, 2, 3], ['a', 'b'])
    res_df = ValDrop([4], 'a').apply(df, verbose=True)
    assert 1 in res_df.index
    assert 2 not in res_df.index
    assert 3 in res_df.index


def test_valdrop_without_columns():
    """Testing the ColDrop pipeline stage."""
    df = pd.DataFrame([[1, 4], [4, 5], [18, 11]], [1, 2, 3], ['a', 'b'])
    res_df = ValDrop([4]).apply(df)
    assert 1 not in res_df.index
    assert 2 not in res_df.index
    assert 3 in res_df.index


def test_valdrop_w_fittable_cq():
    df = pd.DataFrame([[1, 4], [4, 5]], [1, 2], ['aa', 'ba'])
    vdrop = ValDrop([4], columns=StartWith('a'))
    res_df = vdrop(df)
    assert 1 in res_df.index
    assert 2 not in res_df.index
    # now after the column qualifier is fitter, 'ag' should not be transformed
    df = pd.DataFrame([[1, 4], [4, 5]], [1, 2], ['aa', 'ag'])
    res_df = vdrop(df)
    assert 1 in res_df.index
    assert 2 not in res_df.index



"""Testing basic pipeline stages."""

import pandas as pd
import pytest

from pdpipe import DropDuplicates
from pdpipe.exceptions import FailedPreconditionError


def _test_df():
    return pd.DataFrame([[8, 1], [8, 2], [9, 2]], [1, 2, 3], ['a', 'b'])


def test_drop_duplicates():
    df = _test_df()
    stage = DropDuplicates('a')
    res = stage.apply(df)
    assert 1 in res.index
    assert 2 not in res.index
    assert 3 in res.index

    stage = DropDuplicates('b')
    res = stage.apply(df)
    assert 1 in res.index
    assert 2 in res.index
    assert 3 not in res.index

    stage = DropDuplicates(['a', 'b'])
    res = stage.apply(df)
    assert 1 in res.index
    assert 2 in res.index
    assert 3 in res.index

    stage = DropDuplicates()
    res = stage.apply(df, verbose=True)
    assert 1 in res.index
    assert 2 in res.index
    assert 3 in res.index

    stage = DropDuplicates('c')
    with pytest.raises(FailedPreconditionError):
        stage.apply(df)



"""
The `pdpipe` Python package provides a concise interface for building `pandas`
pipelines that have pre-conditions, are verbose, support the fit-transform
design of scikit-learn transformers and are highly serializable. `pdpipe`
pipelines have a simple interface, informative prints and errors on pipeline
application, support pipeline arithmetics and enable easier handling of
mixed-type data.

.. include:: ./documentation.md
"""
# pylint: disable=C0413
# flake8: noqa

import warnings
import traceback


from . import core
from .core import PdPipelineStage, AdHocStage, PdPipeline, make_pdpipeline

core.__load_stage_attributes_from_module__("pdpipe.core")

from . import basic_stages
from .basic_stages import (
    ColDrop,
    ValDrop,
    ValKeep,
    ColRename,
    DropNa,
    SetIndex,
    FreqDrop,
    ColReorder,
    RowDrop,
    Schematize,
    DropDuplicates,
    ColumnDtypeEnforcer,
)

core.__load_stage_attributes_from_module__("pdpipe.basic_stages")

from . import col_generation
from .col_generation import (
    Bin,
    OneHotEncode,
    MapColVals,
    ApplyToRows,
    ApplyByCols,
    ColByFrameFunc,
    AggByCols,
    Log,
)

core.__load_stage_attributes_from_module__("pdpipe.col_generation")

from . import text_stages
from .text_stages import (
    RegexReplace,
    DropTokensByLength,
    DropTokensByList,
)

core.__load_stage_attributes_from_module__("pdpipe.text_stages")

from . import wrappers
from .wrappers import (
    FitOnly,
)

core.__load_stage_attributes_from_module__("pdpipe.wrappers")

try:
    from . import sklearn_stages
    from .sklearn_stages import (
        Encode,
        Scale,
        TfidfVectorizeTokenLists,
    )

    core.__load_stage_attributes_from_module__("pdpipe.sklearn_stages")
except ImportError:
    tb = traceback.format_exc()
    warnings.warn(tb)
    warnings.warn(
        "pdpipe: Scikit-learn or skutil import failed. Scikit-learn"
        "-dependent pipeline stages will not be loaded."
    )

try:
    from . import nltk_stages
    from .nltk_stages import (
        TokenizeText,
        UntokenizeText,
        RemoveStopwords,
        SnowballStem,
        DropRareTokens,
    )

    core.__load_stage_attributes_from_module__("pdpipe.nltk_stages")
except ImportError:
    tb = traceback.format_exc()
    warnings.warn(tb)
    warnings.warn(
        "pdpipe: nltk import failed. nltk-dependent  pipeline "
        "stages will not be loaded."
    )

from . import cq
from . import cond


from ._version import get_versions

__version__ = get_versions()["version"]

for name in [
    "warnings",
    "traceback",
    "_custom_formatwarning",
    "core",
    "basic_stages",
    "sklearn_stages",
    "col_generation",
    "shared",
    "util",
    "_version",
    "get_versions",
]:
    try:
        globals().pop(name)
    except KeyError:
        pass
try:
    del name  # pylint: disable=W0631
except NameError:
    pass

# this dictates which modules are skipped on pdoc documentation generation
__pdoc__ = {
    'shared': False,
}



"""Custom exceptions for pdpipe."""


class FailedPreconditionError(Exception):
    """An exception raised when a pipeline stage is applied to a dataframe for
    which the stage precondition does not hold.
    """


class FailedPostconditionError(Exception):
    """An exception raised when an expected post-condition does not hold after
    and application of the apipine stage to a dataframe.
    """


class PipelineApplicationError(Exception):
    """An exception raised when an exception is thrown during the application
    of a pipeline or a pipeline stage to a dataframe.
    """


class UnfittedPipelineStageError(Exception):
    """An exception raised when a (non-fit) transform is attempted with an
    unfitted pipeline stage.
    """



"""Testing basic pipeline stages."""

import pytest
import pandas as pd

from pdpipe.basic_stages import RowDrop
from pdpipe.exceptions import FailedPreconditionError

DF1 = pd.DataFrame(
    data=[[1, 4], [4, 5], [18, 11]],
    index=[1, 2, 3],
    columns=['a', 'b'],
)

DF2 = pd.DataFrame(
    data=[[1, 4, 3], [4, 5, -3], [18, 11, 9]],
    index=[1, 2, 3],
    columns=['a', 'b', 'c'],
)

DF3 = pd.DataFrame(
    data=[[1, 2], [0, 5], [18, 11]],
    index=[1, 2, 3],
    columns=['a', 'b'],
)


def test_row_drop():
    """Testing the ColDrop pipeline stage."""
    res_df = RowDrop([lambda x: x < 2]).apply(DF1)
    assert 1 not in res_df.index
    assert 2 in res_df.index
    assert 3 in res_df.index


def test_row_drop_verbose():
    """Testing the ColDrop pipeline stage."""
    res_df = RowDrop([lambda x: x < 2]).apply(DF1, verbose=True)
    assert 1 not in res_df.index
    assert 2 in res_df.index
    assert 3 in res_df.index


def test_row_drop_columns():
    """Testing the ColDrop pipeline stage."""
    res_df = RowDrop([lambda x: x < 2]).apply(DF2)
    assert 1 not in res_df.index
    assert 2 not in res_df.index
    assert 3 in res_df.index

    res_df = RowDrop([lambda x: x < 2], columns=['a', 'b']).apply(DF2)
    assert 1 not in res_df.index
    assert 2 in res_df.index
    assert 3 in res_df.index


def test_row_drop_bad_reducer():
    """Testing the ColDrop pipeline stage."""
    with pytest.raises(ValueError):
        RowDrop([lambda x: x < 2], reduce='al')


def test_row_drop_bad_condition_in_dict():
    """Testing the ColDrop pipeline stage."""
    with pytest.raises(ValueError):
        RowDrop({'a': 'bad'})


def test_row_drop_bad_condition_in_list():
    """Testing the ColDrop pipeline stage."""
    with pytest.raises(ValueError):
        RowDrop(['bad'])


def test_row_drop_all_reducer():
    """Testing the ColDrop pipeline stage."""
    res_df = RowDrop([lambda x: x < 3]).apply(DF3)
    assert 1 not in res_df.index
    assert 2 not in res_df.index
    assert 3 in res_df.index

    res_df = RowDrop([lambda x: x < 3], reduce='all').apply(DF3)
    assert 1 not in res_df.index
    assert 2 in res_df.index
    assert 3 in res_df.index


def test_row_droo_bad_columns():
    """Testing the ColDrop pipeline stage."""
    with pytest.raises(FailedPreconditionError):
        RowDrop([lambda x: x < 2], columns=['d']).apply(DF1)


def test_row_drop_xor_reducer():
    """Testing the ColDrop pipeline stage."""
    res_df = RowDrop([lambda x: x < 3]).apply(DF3)
    assert 1 not in res_df.index
    assert 2 not in res_df.index
    assert 3 in res_df.index

    res_df = RowDrop([lambda x: x < 3], reduce='xor').apply(DF3)
    assert 1 in res_df.index
    assert 2 not in res_df.index
    assert 3 in res_df.index
